import numpy as np


def searchsorted2d(ar_a: np.ndarray, ar_b: np.ndarray) -> np.ndarray:
    """
    Inserts ith element of b into sorted ith row of a

    Parameters
    ----------
    ar_a: np.ndarray
            rxc matrix, each rows is sorted
    ar_b: np.ndarray
            rx1 vector

    Returns
    -------
    np.ndarray
            rx1 vector of locations
    """
    dim1_a, dim2_a = ar_a.shape
    ar_b = ar_b.ravel()
    if ar_b.shape[0] != ar_a.shape[0]:
        raise ValueError("Number of values of array b must equal number of rows of array a")
    max_num = np.maximum(ar_a.max() - ar_a.min(), ar_b.max() - ar_b.min()) + 1
    r_seq = max_num * np.arange(ar_a.shape[0])
    indices = np.searchsorted(((ar_a.T + r_seq).T).ravel(), ar_b + r_seq)
    return indices - dim2_a * np.arange(dim1_a)


def shiftarray(arr: np.ndarray, num: int, fill_value: np.float64 = np.nan) -> np.ndarray:
    """Shifts array by num to the right

    Parameters
    ----------
    arr: np.ndarray
        Array to be shifted
    num: int
        Shift size (negative - left shift)
    fill_value: np.float64
        Fill value
    """
    result = np.empty_like(arr)
    if num > 0:
        result[:num] = fill_value
        result[num:] = arr[:-num]
    elif num < 0:
        result[num:] = fill_value
        result[:num] = arr[-num:]
    else:
        result[:] = arr
    return result


def idx_last_nz(inp: np.ndarray | list) -> np.ndarray:
    """Index of the closest previous nonzero element for each element in the array.
    If the array starts with 0 - the index is -1

    Parameters
    ----------
    inp : np.ndarray
        Input array

    Returns
    -------
    np.ndarray
    """
    if not isinstance(inp, np.ndarray):
        inp = np.array(inp)
    nzs = np.concatenate(([-1], np.nonzero(inp)[0]))
    nzcounts = np.cumsum(inp > 0)
    return nzs[nzcounts]


def idx_next_nz(inp: np.ndarray | list) -> np.ndarray:
    """Index of the closest next nonzero element for each element in the array.
    If the array starts with 0 - the index is len(input)

    Parameters
    ----------
    inp : np.ndarray
        Input array

    Returns
    -------
    np.ndarray
    """
    result = idx_last_nz(inp[::-1])
    result = len(inp) - result - 1
    return result[::-1]



import re
from os.path import isfile

import numpy as np
import pyfaidx


def revcomp(seq: str | list | np.ndarray) -> str | list | np.ndarray:
    """Reverse complements DNA given as string

    Parameters
    ----------
    :param: seq Union[str,list,np.ndarray]
        DNA string
    :raises ValueError: is seq is not of the right type

    :return: str | list | np.ndarray


    """
    complement = {
        "A": "T",
        "C": "G",
        "G": "C",
        "T": "A",
        "a": "t",
        "c": "g",
        "g": "c",
        "t": "a",
    }
    if isinstance(seq, str):
        reverse_complement = "".join(complement.get(base, base) for base in reversed(seq))
    elif isinstance(seq, list):
        reverse_complement = [complement.get(base, base) for base in reversed(seq)]
    elif isinstance(seq, np.ndarray):
        reverse_complement = np.array([complement.get(base, base) for base in reversed(seq)])
    else:
        raise ValueError(f"Got unexpected variable {seq} of type {type(seq)}, expected str, list or numpy array")

    return reverse_complement


def hmer_length(seq: pyfaidx.Sequence, start_point: int) -> int:
    """Return length of hmer starting at point start_point

    Parameters
    ----------
    seq: pyfaidx.Sequence
        Sequence
    start_point: int
        Starting point

    Returns
    -------
    int
        Length of hmer (at least 1)
    """

    idx = start_point
    while seq[idx].seq.upper() == seq[start_point].seq.upper():
        idx += 1
    return idx - start_point


def get_chr_sizes(sizes_file: str) -> dict:
    """Returns dictionary from chromosome name to size

    Parameters
    ----------
    sizes_file: str
        .sizes file (use e.g.  cut -f1,2 Homo_sapiens_assembly19.fasta.fai > Homo_sapiens_
        assembly19.fasta.sizes to generate), .fai file or .dict file.
        Any file which doesnot end with fai or dict will be considered .sizes

    Returns
    -------
    dict:
        Dictionary from name to size
    """

    if not isfile(sizes_file):
        raise FileNotFoundError(f"Input_file {sizes_file} not found")
    if sizes_file.endswith("dict"):
        chrom_sizes = {}
        with open(sizes_file, encoding="utf-8") as file:
            for line in file:
                if line.startswith("@SQ"):
                    fields = line[3:].strip().split("\t")
                    row = {}  # Dict of all fields in the row
                    for field in fields:
                        key, value = field.split(":", 1)
                        row[key] = value
                    chrom_sizes[row["SN"]] = int(row["LN"])
        return chrom_sizes
    with open(sizes_file, encoding="ascii") as sizes:
        tmp = [x.strip().split()[:2] for x in sizes]
    return {x[0]: int(x[1]) for x in tmp}


def get_max_softclip_len(cigar):
    group = re.match("(?P<start>[0-9]+S)?[0-9]+[0-9MID]+[MID](?P<end>[0-9]+S)?", cigar).groups()
    start = int(group[0][:-1]) if group[0] else 0
    end = int(group[1][:-1]) if group[1] else 0
    return max(start, end)



#!/bin/env python
import argparse
import pickle
import sys

from ugbio_core.vcfbed.variant_annotation import VcfAnnotator

# Helper script to annotate a single contig of a VCF file


def get_args(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument("--vcf_in", required=True, help="path to input vcf")
    parser.add_argument("--vcf_out", required=True, help="path to otuput vcf")
    parser.add_argument(
        "--annotators_pickle",
        required=True,
        help="path to list of VcfAnnotator pickled classes to annotate with",
    )
    parser.add_argument(
        "--contig",
        required=True,
        help="name of contig to process (this script analyses a contig at a time, "
        "use VcfAnnotator.process_vcf to process multiple contigs)",
    )
    parser.add_argument("--chunk_size", type=int, required=True, help="size of chunks to process in memory")
    return parser.parse_args(argv)


def run(argv):
    """Helper script to annotate a single contig of a VCF file"""
    args = get_args(argv[1:])
    with open(args.annotators_pickle, "rb") as f:
        annotators = pickle.load(f)  # noqa: S301
    VcfAnnotator.process_contig(
        vcf_in=args.vcf_in,
        vcf_out=args.vcf_out,
        annotators=annotators,
        contig=args.contig,
        chunk_size=args.chunk_size,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



# Flow-based read/haplotype class
# This will be a class that would hold a read in flow base
from __future__ import annotations

import copy
import re
from enum import Enum

import numpy as np
import pandas as pd
import pysam
from ugbio_core import array_utils as utils
from ugbio_core import dna_sequence_utils as dnautils
from ugbio_core import math_utils as phred
from ugbio_core.consts import DEFAULT_FLOW_ORDER
from ugbio_core.flow_format import error_model, simulator

DEFAULT_ERROR_MODEL_FN = "error_model.r2d.hd5"
MINIMAL_CALL_PROB = 0.1
DEFAULT_FILLER = 0.0001


def get_bam_header(rgid: str | None = None):
    if rgid is None:
        rgid = "1"
    dct = {
        "HQ": {"VN": "0.1", "SO": "unsorted"},
        "RG": {"ID": rgid, "FO": DEFAULT_FLOW_ORDER, "PL": "ULTIMA", "SM": "NA12878"},
    }

    header = pysam.AlignmentHeader(dct)
    return header


def key2base(key: np.ndarray) -> np.ndarray:
    """
    Returns an array that for every flow outputs the last base output at the beginning of the flow

    Parameters
    ----------
    None

    Returns
    -------
    np.ndarray
            Array of the last output base BEFORE each flow (For instance for flows 01021 the output is -1 -1 0 0 2 )

    """

    n_reps_shifted = utils.shiftarray(key, 1, 0)
    flow2base = -1 + np.cumsum(n_reps_shifted)
    return flow2base


def generate_key_from_sequence(
    sequence: str, flow_order: str, truncate: int | None = None, *, non_standard_as_a: bool = False
) -> np.ndarray:
    """Converts bases to flow order

    Parameters
    ----------
    sequence : str
        Input sequence (bases)
    flow_order : str
        Flow order
    truncate : int, optional
        maximal hmer to read
    non_standard_as_a: bool, optional
        Replace non-standard nucleotides with A (default: false)

    Returns
    -------
    np.ndarray
        sequence in key space

    Raises
    ------
    ValueError
        When  there are non-standard nucleotides
    """
    # sanitize input
    sequence = sequence.upper()
    if bool(re.compile(r"[^ACGT]").search(sequence)):
        if non_standard_as_a:
            sequence = re.sub(r"[^ACGT]", "A", sequence)
        else:
            raise ValueError(
                "Input contains non ACGTacgt characters" + (f":\n{sequence}" if len(sequence) <= 100 else "")  # noqa: PLR2004
            )

    # process
    flow = flow_order * len(sequence)

    key = []
    pos = 0
    for base in flow:
        hcount = 0
        for i in range(pos, len(sequence)):
            if sequence[i] == base:
                hcount += 1
            else:
                break
        else:
            key.append(hcount)
            break  # end of sequence

        key.append(hcount)
        pos += hcount
    if truncate:
        return np.clip(np.array(key), 0, truncate)
    return np.array(key)


class SupportedFormats(Enum):
    MATT = "matt"
    ILYA = "ilya"
    CRAM = "cram"


class FlowBasedRead:
    """Class that helps working with flow based reads

    This is the class that allows for comparison between
    reads in flow base, applying error model, outputting uBAM etc.

    Attributes
    ----------
    read_name: str
        Read name
    seq: str
        Original read sequence
    key: np.ndarray
        sequence in flow base
    flow2base: np.ndarray
        position of the last output sequence base _before_ each flow for forward key
    flow_order: str
        sequence of flows
    _motif_size: int
        size of motif (left or right)
    _regressed_signal: np.ndarray
        Regressed signal array **binned relative to the error model**
    _max_hmer: int
        Maximal hmer
    base_to_flow_mapping: np.ndarray
        Map read base to the number of flow it was read
    Methods
    -------
    apply_cigar:
        Returns a new read with cigar applied (takes care of hard clipping and soft clipping)
    """

    key: np.ndarray
    flow_order: str
    cigar: list
    _max_hmer: int
    seq: str
    read_name: str
    is_reverse: bool
    start: int
    end: int
    flow2base: np.ndarray
    _motif_size: int
    _regressed_signal: np.ndarray
    base_to_flow_mapping: np.ndarray

    def __init__(self, dct: dict):
        """Generic constructor

        Parameters
        ----------
        Receives key-value dictionary and sets the corresponding properties of the object

        Returns
        -------
        None
        """

        for k in dct:
            setattr(self, k, dct[k])
        if not hasattr(self, "key"):
            raise AttributeError("Something is broken in the constructor, key is not defined")
        if not hasattr(self, "flow_order"):
            raise AttributeError("Something is broken in the constructor, flow_order is not defined")
        self.flow2base = key2base(self.key).astype(int)
        self.flow_order = simulator.get_flow2_base(self.flow_order, len(self.key))

        self._validate_seq()

    @classmethod
    def from_tuple(
        cls,
        read_name: str,
        read: str,
        this_error_model: error_model.ErrorModel = None,
        flow_order: str = DEFAULT_FLOW_ORDER,
        motif_size: int = 5,
        max_hmer_size: int = 9,
    ):
        """Constructor from FASTA record and error model. Sets `seq`, `r_seq`, `key`,
        `rkey`, `flow_order`, `r_flow_order` attributes.

        Parameters
        ----------
        read_name: str
            Name of the read
        read: str
            String of bases
        this_error_model: pd.DataFrame
            Error model from motif, hmer to probability that data with +1,-1,0 of that hmer
            generated this motif.
            Error model could be None, in this case it will be assumed that there are no errors
        flow_order: str
            The flow cycle
        motif_size: int
            Size of the motif
        max_hmer_size: int
            Maximal size of the called hmer
        """
        dct = {}
        dct["read_name"] = read_name
        dct["seq"] = read
        dct["forward_seq"] = read
        dct["key"] = generate_key_from_sequence(dct["forward_seq"], flow_order=flow_order)
        dct["flow_order"] = flow_order
        dct["_error_model"] = this_error_model
        dct["_max_hmer"] = max_hmer_size
        dct["_motif_size"] = motif_size
        dct["direction"] = "synthesis"
        return cls(dct)

    @classmethod
    def from_sam_record(  # noqa: C901, PLR0912, PLR0915 #TODO: refactor
        cls,
        sam_record: pysam.AlignedSegment,
        this_error_model: error_model.ErrorModel | None = None,
        flow_order: str = DEFAULT_FLOW_ORDER,
        motif_size: int = 5,
        max_hmer_size: int = 12,
        filler: float = DEFAULT_FILLER,
        min_call_prob: float = MINIMAL_CALL_PROB,
        _fmt: str = "cram",
        *,
        spread_edge_probs=True,
        validate=False,
    ):
        """Constructor from BAM record and error model. Sets `seq`, `r_seq`, `key`,
        `rkey`, `flow_order`, `r_flow_order` and `_flow_matrix` attributes

        Parameters
        ----------
        read_name: str
            Name of the read
        seq: str
            DNA sequence of the read (basecalling output)
        flow_order: np.ndarray
            Array of chars - base for each flow
        this_error_model: pd.DataFrame
            Error model from motif, hmer to probability that data with +1,-1,0 of that hmer generated this motif
            Can be optional if the read has ks, kq, kd fields
        motif_size: int
            Size of the motif
        max_hmer_size: int
            Maximal reported hmer size
        filler: float
            The minimal probability to appear in the flow flow_matrix, will be attempted to be quessed from the master,
            if not the default will be used (default: %f)
        min_call_prob: float
            The minimal probability to be placed on the call (default: %f)
        _fmt: str
            Can be 'matt', 'ilya' or 'cram' (the current BARC output format)
        spread_edge_probs: bool
            Should the error probabilities of the edge flows be smoothed (e.g. after trimming). Default - true
        validate: bool
            Will print number of illegal flows (in the tp). Default - false

        Returns
        -------
        Object
        """
        dct = {}
        dct["record"] = sam_record
        dct["read_name"] = sam_record.query_name
        dct["seq"] = sam_record.query_sequence
        dct["is_reverse"] = sam_record.is_reverse
        if sam_record.is_reverse:
            dct["forward_seq"] = dnautils.revcomp(dct["seq"])
        else:
            dct["forward_seq"] = dct["seq"]
        dct["_error_model"] = this_error_model

        try:
            fmt: SupportedFormats = SupportedFormats[_fmt.upper()]
        except KeyError as err:
            raise RuntimeError(f"format {_fmt} not supported") from err

        if fmt == SupportedFormats.ILYA:
            if sam_record.has_tag("ks"):
                dct["key"] = np.array(sam_record.get_tag("ks"), dtype=np.int8)
            elif sam_record.has_tag("kr"):
                dct["key"] = np.array(sam_record.get_tag("kr"), dtype=np.int8)
            else:
                dct["key"] = generate_key_from_sequence(dct["forward_seq"], flow_order=flow_order)
        elif fmt == SupportedFormats.MATT:
            dct["key"] = np.array(sam_record.get_tag("kr"), dtype=np.int8)
        elif fmt == SupportedFormats.CRAM:
            dct["key"] = generate_key_from_sequence(dct["seq"], flow_order=flow_order)

        dct["_max_hmer"] = max_hmer_size
        dct["_motif_size"] = motif_size
        dct["flow_order"] = flow_order
        if fmt == SupportedFormats.ILYA:
            if sam_record.has_tag("kf"):
                row = sam_record.get_tag("kh")
                col = sam_record.get_tag("kf")
                vals = sam_record.get_tag("kd")
                shape = (max_hmer_size + 1, len(dct["key"]))
                flow_matrix = cls._matrix_from_sparse(row, col, vals, shape, DEFAULT_FILLER)
                dct["_flow_matrix"] = flow_matrix
        elif fmt == SupportedFormats.MATT:
            row = np.array(sam_record.get_tag("kh"))
            col = np.array(sam_record.get_tag("kf"))
            vals = np.array(sam_record.get_tag("kd"))
            row = np.concatenate((row, dct["key"].astype(row.dtype)))
            col = np.concatenate((col, np.arange(len(dct["key"])).astype(col.dtype)))
            vals = np.concatenate((vals, np.zeros(len(dct["key"]), dtype=float)))
            shape = (max_hmer_size + 1, len(dct["key"]))
            flow_matrix = cls._matrix_from_sparse(row, col, vals, shape, DEFAULT_FILLER)
            dct["_flow_matrix"] = flow_matrix
        elif fmt == SupportedFormats.CRAM:
            if sam_record.has_tag("t0"):
                t0 = sam_record.get_tag("t0")
            else:
                t0 = None
            if validate:
                print(str(sam_record.query_name) + ":", end=" ")
            flow_matrix = cls._matrix_from_qual_tp(
                dct["key"],
                np.array(sam_record.query_qualities, dtype=int),
                tp_tag=np.array(sam_record.get_tag("tp"), dtype=int),
                t0_tag=t0,
                fallback_filler=filler,
                min_call_prob=min_call_prob,
                max_hmer_size=max_hmer_size,
                spread_edge_probs=spread_edge_probs,
                validate=validate,
            )
            dct["_flow_matrix"] = flow_matrix

        dct["cigar"] = sam_record.cigartuples
        dct["start"] = sam_record.reference_start
        dct["end"] = sam_record.reference_end
        if fmt != SupportedFormats.CRAM:
            dct["direction"] = "synthesis"
        else:
            dct["direction"] = "reference"
        return cls(dct)

    @classmethod
    def _matrix_from_qual_tp(
        cls,
        key: np.ndarray,
        qual: np.ndarray,
        tp_tag: np.ndarray,
        t0_tag: np.ndarray | None = None,
        fallback_filler: float = DEFAULT_FILLER,
        min_call_prob: float = MINIMAL_CALL_PROB,
        max_hmer_size: int = 12,
        *,
        spread_edge_probs: bool = True,
        validate: bool = False,
    ) -> np.ndarray:
        """Fill flow matrix from the CRAM format data. Here is the description

        To fill the probability (P(i,j)) for call i for flow j
        QUAL values of the homopolymer output at flow j correspond to the probabilities
        of the errors that are defined by the tp tag.

        Example:

        Sequence  AAAAA
        QUAL      BCICB
        tp        -1, 1, 0, 1, -1

        This correspond to the probabilities of 4-mer (5-mer -1) and 6-mer (5-mer + 1)
        To calculate:
        P(4) = 2*PhredToProb('B') = 0.001,
        P(6) = 2*PhredToProb('C') = 0.0008,
        P(5) = max(1-sum(P(i)) for i != 5

        The values that are not mentioned are filled with `filler` value

        For the flows where the call is zero all the non-zero probabilities are initialized with filler
        If optional t0 tag is given, for the zero calls the P(1) = min(P_{left}, P_{right} ) where
        P_{left} and P_{right} are the qualities in t0 tag of the two non-zero neighboring flows

        See https://ultimagen.atlassian.net/wiki/spaces/AG/pages/1797324911/SAM+Format for more details.


        Parameters
        ----------
        key : np.ndarray
            Flow-space sequence (starting from the first flow of the flow order)
        qual : np.ndarray
            Quality array of the read
        tp_tag : np.ndarray
            tp tag of the read
        t0_tag : np.ndarray
            T0 tag of the read, optional
        fallback_filler : float
            Value to place as minimal probability, only in cases not being able to be guessed from the read
        min_call_prob: float
            The minimal value to place as the probability of the actual call
        max_hmer_size : int, optional
            Maximum hmer size that has a reported probability
        spread_edge_probs: bool, optional
            Should the error probabilities of the edge bases be smoothed (e.g. after trimming). Default - true
        validate: bool, optional
            Should the function validate the input. Default - false

        Returns
        -------
        np.ndarray
            max_hmer+1 x n_flows flow matrix
        """

        filler = cls._estimate_filler_value(qual, tp_tag, fallback_filler) / max_hmer_size
        flow_matrix = np.ones((max_hmer_size + 1, len(key))) * filler

        probs = phred.unphred(qual)
        if t0_tag is not None:
            t0_probs = phred.unphred_str(t0_tag)

        # base_to_flow_index[i] is the flow number which read base i
        base_to_flow_index = np.repeat(np.arange(len(key)), key.astype(int))
        if t0_tag is not None:
            t0_on_flows = np.zeros(len(key))
            t0_on_flows[base_to_flow_index] = t0_probs
            left_neighbor = utils.idx_last_nz(key)
            right_neighbor = utils.idx_next_nz(key)
            left_prob = t0_on_flows[np.clip(left_neighbor, 0, None)]
            left_prob[left_neighbor < 0] = 0
            right_prob = t0_on_flows[np.clip(right_neighbor, None, len(key) - 1)]
            right_prob[right_neighbor >= len(key)] = 0
            flow_matrix[1, key == 0] = np.min(np.vstack((left_prob, right_prob)), axis=0)[key == 0]
            flow_matrix[1, (flow_matrix[1, :] / max_hmer_size <= 3 * filler) & (key == 0)] = filler
        # repeat each value k (hmer-size) k times, e.g [0, 1, 2, 1, 3] -> [1, 2, 2, 1, 3, 3, 3]
        # This results in hmer_sizes in base-space
        # Notice a trimmed hmer will result in the wrong original hmer-size
        hmer_sizes = np.repeat(key, key.astype(int))

        # place_to_locate[i] is the hmer-size which the probability in base[i] refers to
        # This size will be used to locate the probability in the flow-matrix
        place_to_locate = tp_tag + hmer_sizes

        place_to_locate = np.clip(place_to_locate, None, max_hmer_size)
        if spread_edge_probs:
            first_call = key[np.nonzero(key)[0][0]]
            place_to_locate[:first_call] = np.clip(place_to_locate[:first_call], 0, None)
            last_call = key[np.nonzero(key)[0][-1]]
            place_to_locate[last_call - 1 :] = np.clip(place_to_locate[last_call - 1 :], 0, None)
            # place_to_locate[0] = np.clip(place_to_locate[0], 0, None)
            # place_to_locate[-1] = np.clip(place_to_locate[-1], 0, None)

        if not np.all(place_to_locate >= 0):
            raise ValueError("Wrong position to place")
        flat_loc = np.ravel_multi_index((place_to_locate, base_to_flow_index), flow_matrix.shape)
        out = np.bincount(flat_loc, probs)
        flow_matrix.flat[flat_loc] = out[flat_loc]

        flow_matrix[np.clip(key, None, max_hmer_size), np.arange(flow_matrix.shape[1])] = 0
        total = np.sum(flow_matrix, axis=0)

        diff = np.clip(1 - total, min_call_prob, None)
        flow_matrix[np.clip(key, None, max_hmer_size), np.arange(flow_matrix.shape[1])] = diff
        if spread_edge_probs:
            first_call = np.nonzero(key)[0][0]
            flow_matrix[:, first_call] = 1 / (max_hmer_size + 1)
            last_call = np.nonzero(key)[0][-1]
            flow_matrix[:, last_call] = 1 / (max_hmer_size + 1)
        if validate:
            violations_tp = 0
            violations_qual = 0
            start_locations = np.cumsum(key)
            for i in start_locations[1:-2]:
                tp_take = tp_tag[i : i + hmer_sizes[i]]
                violations_tp += (tp_take != tp_take[::-1]).any()
                qual_take = qual[i : i + hmer_sizes[i]]
                violations_qual += (qual_take != qual_take[::-1]).any()

            print(f"Number of tp symmetry violations: {violations_tp}, qual symmetry violations: {violations_qual}")
        return flow_matrix

    @classmethod
    def from_sam_record_rsig(
        cls,
        sam_record: pysam.AlignedSegment,
        regressed_signal=np.ndarray,
        this_error_model: error_model.ErrorModel | None = None,
        flow_order: str = DEFAULT_FLOW_ORDER,
        motif_size: int = 5,
        max_hmer_size: int = 9,
    ):
        """Constructor from BAM record and error model. Sets `seq`, `r_seq`, `key`,
        `rkey`, `flow_order`, `r_flow_order` and `_flow_matrix` attributes

        Parameters
        ----------
        sam_record: pysam.AlignedSegment
            `The Sam record
        regressed_signal: np.ndarray
            `Array of regressed signal
        this_error_model: error_model.ErrorModel, optional
            The error model
        flow_order: str
            Flow order - array of chars - base for each flow
        motif_size: int
            Size of the motif
        max_hmer_size: int
            Maximal reported hmer size

        Returns
        -------
        Object
        """
        dct = vars(cls.from_sam_record(sam_record, this_error_model, flow_order, motif_size, max_hmer_size))

        dct["_regressed_signal"] = regressed_signal[: len(dct["key"])]
        return cls(dct)

    @classmethod
    def from_sam_record_flow_matrix(
        cls,
        sam_record: pysam.AlignedSegment,
        flow_matrix=np.ndarray,
        flow_order: str = DEFAULT_FLOW_ORDER,
        motif_size: int = 5,
        max_hmer_size: int = 9,
    ):
        """Constructor from BAM record and error model. Sets `seq`, `r_seq`, `key`,
        `rkey`, `flow_order`, `r_flow_order` and `_flow_matrix` attributes

        Parameters
        ----------
        sam_record: pysam.AlignedSegment
            `The Sam record
        flow_matrix: np.ndarray
            Matrix of probabilities
        flow_order: str, optional
            Flow order - array of chars - base for each flow
        motif_size: int, optional
            Size of the motif
        max_hmer_size: int, optional
            Maximal reported hmer size

        Returns
        -------
        Object
        """
        dct = vars(cls.from_sam_record(sam_record, None, flow_order, motif_size, max_hmer_size))

        dct["_flow_matrix"] = flow_matrix[:, : len(dct["key"])]
        return cls(dct)

    def _validate_seq(self) -> None:
        """Validates that there are no hmers longer than _max_hmer

        Parameters
        ----------
        None

        Returns
        -------
        None. Sets attribute _validate
        """
        self._validate = ~np.any(self.key > (self._max_hmer - 1))

    @classmethod
    def _estimate_filler_value(cls, qual: np.ndarray, tp_tag: np.ndarray, fallback_filler: float) -> float:
        """Estimates the maximal reported quality from the read qualities

        Parameters
        ----------
        qual : np.ndarray
            Quality array of the read
        tp_tag : np.ndarray
            tp tag of the read
        fallback_filler : float
            If there are no zero calls, this value will be used

        Returns
        -------
        float
            The estimated filler value
        """
        if np.sum(tp_tag == 0) == 0:
            return fallback_filler
        max_qual: float = np.max(qual[tp_tag == 0])
        return float(phred.unphred(max_qual))

    def is_valid(self) -> bool:
        """Returns if the key is valid"""
        return self._validate

    # TODO: no imports, can be deleted?
    def read2_flow_matrix(self, *, regressed_signal_only: bool = False) -> tuple:
        """Gets the hmerxflow probability matrix
        matrix[i,j] = P(read_hmer==read_hmer[j] | data_hmer[j]==i)

        Paramters
        ---------
        regressed_signal_only: bool
            Use only regressed_signal without motifs
        Returns
        -------
        np.ndarray
            Flow matrix of (max_hmer_size+1) x n_flow

        """
        if not hasattr(self, "_flow_matrix"):
            if regressed_signal_only:
                self._flow_matrix = self._get_single_flow_matrix_only_regressed(self._regressed_signal)
                return self._flow_matrix
            self._flow_matrix = self._get_single_flow_matrix(self.key, self.flow2base, self.forward_seq)
        return self._flow_matrix

    def _get_single_flow_matrix_only_regressed(self, regressed_signal: np.ndarray, threshold=0.01):
        probs1 = 1 / np.add.outer(regressed_signal, -np.arange(self._max_hmer + 1)) ** 2
        probs1 /= probs1.sum(axis=-1, keepdims=True)

        if threshold is not None:
            probs1[probs1 < threshold] = 0
        probs1 = self._fix_nan(probs1)
        return probs1.T

    def _get_single_flow_matrix(
        self,
        key: np.ndarray,
        flow2base: np.ndarray,
        seq: str,
    ) -> np.ndarray:
        """Returns flow matrix column for a given flow key. Note that if the error model is None
        it is assumed that there are no errors (useful for getting flow matrix of a haplotype)

        Parameters
        ----------
        key: np.ndarray
            Hmer for each flow
        flow2base : np.ndarray
            For each flow - what was the last base output **before** the flow
        seq: str
            Sequence of the read

        Returns
        -------
        hmerxflow probability matrix
        """
        if self._error_model is None:
            flow_matrix = np.zeros((self._max_hmer + 1, len(key)))
            flow_matrix[np.clip(self.key, 0, self._max_hmer), np.arange(len(self.key))] = 1
            return flow_matrix

        motifs_left = []
        for i in range(key.shape[0]):
            left_base = flow2base[i]
            if left_base < 0:
                motifs_left.append("")
            else:
                motifs_left.append(seq[max(left_base - self._motif_size + 1, 0) : left_base + 1])
                if seq[left_base] == self.flow_order[i]:
                    raise ValueError("Something wrong with motifs")
        motifs_right = []
        for i in range(key.shape[0] - 1):
            right_base = flow2base[i + 1]
            motifs_right.append(seq[right_base + 1 : right_base + self._motif_size + 1])
            if seq[right_base + 1] == self.flow_order[i]:
                raise ValueError("Something wrong with motifs")
        motifs_right.append("")
        index = list(zip(motifs_left, key, self.flow_order[: len(key)], motifs_right, strict=False))
        hash_idx = [hash(x) for x in index]

        if hasattr(self, "_regressed_signal"):
            bins = self._regressed_signal[: len(hash_idx)]
        else:
            bins = None
        idx_list = np.array(self._error_model.hash2idx(hash_idx, bins))
        tmp = self._error_model.get_index(idx_list, bins)

        # index
        pos = np.arange(len(key))
        diffs = [key + x for x in np.arange(-(tmp.shape[1] - 1) / 2, (tmp.shape[1] - 1) / 2 + 1)]

        flow_matrix = np.zeros((self._max_hmer + 1, len(key)))

        conc_diffs = np.concatenate(diffs).astype(int)
        position_matrix = np.tile(pos, tmp.shape[1])
        flat_position_matrix = tmp.T.ravel()
        take = (conc_diffs >= 0) & (conc_diffs <= self._max_hmer)

        flow_matrix[conc_diffs[take], position_matrix[take]] = flat_position_matrix[take]
        flow_matrix = self._fix_nan(flow_matrix)
        return flow_matrix

    def _fix_nan(self, flow_matrix: np.ndarray) -> np.ndarray:
        """Fixes cases when there is nan in the flow matrix.
        This is an ugly hack - we assume that nan come from cases when
        there is not enough data - currently we will them by 0.9 for P(R=i|H=i) and
        by 0.1 P(r=i+-1|H=i). The only exception is P(R=i|H=0) these missing data tend to
        come from the fact that such an insertion would require cycle skip that we do not allow

        Parameters
        ----------
        flow_matrix: np.ndarray
            Uncorrected flow matrix

        Returns
        -------
        np.ndarray
            Corrected matrix
        """
        gtr = self.key
        row, col = np.nonzero(np.isnan(flow_matrix))
        take = row == gtr[col]
        zeros = row == 0
        flow_matrix[row[take], col[take]] = 0.9
        flow_matrix[row[~take], col[~take]] = 0.1
        # missing data from P(|H=0) is coming from cycle skips
        flow_matrix[row[zeros], col[zeros]] = 0
        return flow_matrix

    def _get_phred_flow(self) -> np.ndarray:
        """
        Returns Quality of flow in Phred

        Parameters
        ----------
        None

        Returns
        -------
        np.ndarray
            Phred scaled **integer** qualities
        """

        tmp = np.max(self._flow_matrix, axis=0)
        high_conf = tmp == 1
        tmp = -10 * np.log10(1 - tmp)
        tmp[high_conf] = 60
        return tmp.astype(np.int8)

    def _matrix_to_sparse(self, probability_threshold: float = 0) -> tuple:
        """Converts the flow matrix to the tag representation

        Parameters
        ----------
        probability_threshold: float
            Optional threshold that would suppress variants with ratio to the best variant below the threshold

        Returns
        -------
        tuple (np.ndarray, np.ndarray, np.ndarray):
            row, column, values (differences in int(log10) from the maximal value)
        """

        probability_threshold = -10 * np.log10(probability_threshold)
        tmp_matrix = self._flow_matrix.copy()
        tmp_matrix[
            self.key[self.key <= self._max_hmer],
            np.arange(len(self.key))[self.key <= self._max_hmer],
        ] = 0

        row, column = np.nonzero(tmp_matrix)
        values = tmp_matrix[row, column]

        values = np.log10(values)
        col_max = tmp_matrix[np.clip(self.key, 0, self._max_hmer), np.arange(len(self.key))]
        normalized_values = -10 * (values - col_max[column])
        normalized_values = np.clip(normalized_values, -60, 60)

        suppress = normalized_values > probability_threshold

        # do not output the key itself as it is always zero
        suppress = suppress | (self.key[column] == row)
        return row[~suppress], column[~suppress], normalized_values[~suppress]

    @classmethod
    def _matrix_from_sparse(cls, row, column, values, shape, filler):
        flow_matrix = np.ones(shape) * filler
        kd_vals = np.array(values, dtype=float)
        kd_vals = -kd_vals
        kd_vals = kd_vals / 10
        kd_vals = 10 ** (kd_vals)
        flow_matrix[row, column] = kd_vals
        return flow_matrix

    def to_record(
        self,
        hdr: pysam.AlignmentHeader | None = None,
        probability_threshold: float = 0,
    ) -> pysam.AlignedSegment:
        """Converts flowBasedRead into BAM record

        Parameters
        ----------
        hdr: pysam.AlignmentHeader
            Optional header
        probability_threshold : float
            Optional - do not report variants with probability ratio to the best variant
            less than probability_threshold (default: 0)
        Returns
        -------
        pysam.AlignedSegment
        """
        if hasattr(self, "record"):
            res = self.record
        else:
            res = pysam.AlignedSegment(hdr)
            res.query_sequence = self.seq
            res.query_name = self.read_name
        res.set_tag("KS", "".join(self.flow_order[:4]))
        if hasattr(self, "_flow_matrix"):
            alt_row, alt_col, alt_val = self._matrix_to_sparse(probability_threshold=probability_threshold)

            res.set_tag("kr", [int(x) for x in self.key])
            res.set_tag("kh", [int(x) for x in alt_row])
            res.set_tag("kf", [int(x) for x in alt_col])
            res.set_tag("kd", [int(x) for x in alt_val])
        else:
            res.set_tag("kr", [int(x) for x in self.key])
        self.record = res
        return res

    def flow_matrix_to_list(self) -> list:
        """Returns flow matrix as a list of tuples

        Parameters
        ----------
        None

        Returns
        -------
        list
            List of tuples ordered by flow index then hmer
        """
        idcs = np.unravel_index(np.arange(self._flow_matrix.size), self._flow_matrix.T.shape)
        tups = [tuple(x) for x in np.vstack((idcs, np.ravel(self._flow_matrix.T))).T]
        return list(tups)

    def _left_clipped_flows(self, cigar: list | None = None) -> tuple:
        """Returns number of flows clipped from the left

        Parameters
        ----------
        cigar: list
            Optional cigar, if None - the object's cigar will be used

        Returns
        -------
        tuple (int, int)
            (fn, n) subtract n from the hmer count of flow fn, trim all flows before fn
        """
        if cigar is None:
            cigar = self.cigar
        if cigar[0][0] != 5:  # noqa: PLR2004
            return (0, 0)
        bases_clipped = cigar[0][1]
        stop_clip = np.argmax(self.flow2base + self.key >= bases_clipped)
        stop_clip_flow = stop_clip
        hmer_clipped = bases_clipped - self.flow2base[stop_clip] - 1

        return (stop_clip_flow, hmer_clipped)

    def _right_clipped_flows(self, cigar: list | None = None) -> tuple:
        """Returns number of flows clipped from the right

        Parameters
        ----------
        cigar: list
            Optional cigar, if None - the object's cigar will be used

        Returns
        -------
        tuple (int, int)
            (fn, n) subtract n from the hmer count of flow -fn-1, trim all flows after -fn-1
        """
        if cigar is None:
            cigar = self.cigar

        if cigar[-1][0] != 5:  # noqa: PLR2004
            return (0, 0)
        bases_clipped = cigar[-1][1]
        reverse_flow2base = key2base(self.key[::-1])

        stop_clip = np.argmax(reverse_flow2base + self.key[::-1] >= bases_clipped)
        hmer_clipped = bases_clipped - reverse_flow2base[stop_clip] - 1
        return (stop_clip, hmer_clipped)

    def apply_alignment(self) -> FlowBasedRead:  # noqa: C901, PLR0912 #TODO: refactor
        """Applies alignment (inversion / hard clipping ) to the flowBasedRead

        Parameters
        ----------
        None

        Returns
        -------
        New FlowBasedRead with
        Modifies `key`, `_flow_matrix`, `flow2base`, `flow_order` attributes
        """
        if not hasattr(self, "cigar"):
            raise AttributeError("Only aligned read can be modified")
        attrs_dict = vars(self)
        other = FlowBasedRead(copy.deepcopy(attrs_dict))
        if other.is_reverse and self.direction != "reference":
            if hasattr(other, "_flow_matrix"):
                other._flow_matrix = other._flow_matrix[:, ::-1]

            other.key = other.key[::-1]
            other.flow2base = key2base(other.key)
            other.flow_order = dnautils.revcomp(other.flow_order)

            if hasattr(other, "_regressed_signal"):
                other._regressed_signal = other._regressed_signal[::-1]

        clip_left, left_hmer_clip = other._left_clipped_flows()
        clip_right, right_hmer_clip = other._right_clipped_flows()

        if not (left_hmer_clip >= 0 and right_hmer_clip >= 0):
            raise ValueError("Some problem with hmer clips")
        original_length = len(other.key)
        other.key[clip_left] -= left_hmer_clip
        if hasattr(other, "_regressed_signal"):
            other._regressed_signal[clip_left] -= left_hmer_clip
        # if no flows left on the left hmer - truncate it too

        shift_left = True
        if clip_left >= 0 or left_hmer_clip >= 0:
            while other.key[clip_left] == 0:
                clip_left += 1
                shift_left = False

        other.key[-1 - clip_right] -= right_hmer_clip
        if hasattr(other, "_regressed_signal"):
            other._regressed_signal[-1 - clip_right] -= left_hmer_clip

        shift_right = True
        if clip_right >= 0 or right_hmer_clip >= 0:
            while other.key[-1 - clip_right] == 0:
                shift_right = False
                clip_right += 1

        other.key = other.key[clip_left : original_length - clip_right]
        if hasattr(other, "_regressed_signal"):
            other._regressed_signal = other._regressed_signal[clip_left : original_length - clip_right]
        other.flow2base = other.flow2base[clip_left : original_length - clip_right]
        other.flow_order = other.flow_order[clip_left : original_length - clip_right]

        if hasattr(other, "_flow_matrix"):
            other._flow_matrix = other._flow_matrix[:, clip_left : original_length - clip_right]

            if shift_left:
                other._flow_matrix[:, 0] = utils.shiftarray(other._flow_matrix[:, 0], -left_hmer_clip, 0)
            if shift_right:
                other._flow_matrix[:, -1] = utils.shiftarray(other._flow_matrix[:, -1], -right_hmer_clip, 0)
        other.direction = "reference"
        return other

    def _key2str(self):
        return "".join(np.repeat(self.flow_order, self.key))

    def haplotype_matching(self, read: FlowBasedRead) -> np.float64:  # noqa: C901 #TODO: refactor. too complex
        """Returns log likelihood for the match of the read and the haplotype.
        Both the haplotype and the read flow matrices should be aligned in the reference direction.
        It is assumed that haplotype does not have probabilities.

        Parameters
        ----------
            read: FlowBasedRead
                flow based read **after `apply_alignment`**.
        Returns
        -------
        np.float64
            Log likelihood of the match
        """
        if not (self.direction == read.direction and self.direction == "reference"):
            raise ValueError("Only reads aligned to the reference please")

        if not (self.start <= read.start and self.end >= read.end):
            raise ValueError("Read alignment should be contained in the haplotype")

        hap_locs = np.array([x[0] for x in self.record.get_aligned_pairs(matches_only=True)])
        ref_locs = np.array([x[1] for x in self.record.get_aligned_pairs(matches_only=True)])
        hap_start_loc = hap_locs[np.searchsorted(ref_locs, read.start)]
        hap_end_loc = hap_locs[np.searchsorted(ref_locs, read.end - 1)]

        left_clip = hap_start_loc
        right_clip = len(self.seq) - hap_end_loc - 1
        right_clip = max(0, right_clip)

        clipping = []
        if left_clip > 0:
            clipping.append((5, left_clip))
        clipping.append((0, len(read.seq)))
        if right_clip > 0:
            clipping.append((5, right_clip))

        if left_clip < 0 or right_clip < 0 or left_clip >= len(self.seq) or right_clip >= len(self.seq):
            return 1

        if self.key.max() >= 9:  # noqa: PLR2004
            return -np.inf

        clip_left, left_hmer_clip = self._left_clipped_flows(clipping)
        clip_right, right_hmer_clip = self._right_clipped_flows(clipping)

        if abs(left_hmer_clip) >= 11 or abs(right_hmer_clip) >= 11:  # noqa: PLR2004
            raise ValueError("Weird hmer_clip")
        if clip_left >= len(self.key) or clip_right >= len(self.key):
            return -np.inf

        if not (left_hmer_clip >= 0 and right_hmer_clip >= 0):
            raise ValueError("Some problem with hmer clips")
        key = self.key.copy()
        original_length = len(key)
        clip_left = max(clip_left - 4, 0)
        clip_right = min(original_length, original_length - clip_right + 4)
        key = key[clip_left:clip_right]
        flow_order = self.flow_order[clip_left:clip_right]
        starting_points = np.nonzero(flow_order == read.flow_order[0])[0]
        starting_points = starting_points[starting_points + len(read.key) <= len(key)]
        best_alignment = -np.inf
        for start_point in starting_points:
            fetch = np.log10(
                read._flow_matrix[
                    np.clip(
                        key[start_point : start_point + len(read.key)],
                        None,
                        read._flow_matrix.shape[0] - 1,
                    ),
                    np.arange(len(read.key)),
                ]
            )[1:-1].sum()
            best_alignment = max(fetch, best_alignment)
        return best_alignment

    def base_to_flow(self, base: int) -> int:
        """Returns the flow at which the read base base is sequenced

        Parameters
         ----------
         base: int
             Number of the read base

         Returns
         -------
         int
             Flow of the base
        """
        if hasattr(self, "base_to_flow_mapping"):
            return self.base_to_flow_mapping[base]
        self.base_to_flow_mapping = np.repeat(np.arange(len(self.key)), self.key.astype(int))
        return self.base_to_flow_mapping[base]

    def get_flow_matrix_column_for_base(self, base: int) -> tuple[str, np.ndarray]:
        """Returns the flow matrix column for the base

        Parameters
        ----------
        base: int
            Number of the read base

        Returns
        -------
        tuple[str,np.ndarray]
            The flow nucleotide and the flow matrix column
        """
        return self.flow_order[self.base_to_flow(base)], self._flow_matrix[:, self.base_to_flow(base)]


def get_haplotype_by_read_matrix(haplotypes: list, reads: list) -> np.ndarray:
    """Matrix of likelihoods of reads from each haplotype

    Parameters
    ----------
    haplotypes: list
        List of haplotypes (n_hap FlowBasedReads)
    reads: list
        List of reads (n_reads FlowBasedReads)

    Returns
    -------
    n_hapxn_reads matrix of likelihood log(P(read|hap)
    """

    n_hap = len(haplotypes)
    n_reads = len(reads)
    result = np.zeros((n_hap, n_reads))
    for i in range(n_hap):
        for j in range(n_reads):
            result[i, j] = haplotypes[i].haplotype_matching(reads[j])
    return result


def _extract_location(block: str) -> str:
    """Extracts locus from the string

    Parameters
    ----------
    block: str
        Whole block extracted by `parse_haplotype_matrix_file`

    Returns
    -------
    str: chrom:start-end
    """

    return block.split("\n")[0].strip()


def _parse_active_region(block: str) -> list:
    """Parses single haplotype,read, matrix block and returns a
    list of probability matrices by sample

    Parameters
    ----------
    block: str
        Parses a block from the output of HaplotypeCaller

    Returns
    -------
    list
    """
    haplotype_start = block.index(">> Haplotypes")
    haplotype_end = block.index(">> Sample")
    haplotypes = _parse_haplotypes(block[haplotype_start:haplotype_end])
    sample_starts = [x.start() for x in re.finditer(r">> Sample", block)]
    sample_ends = sample_starts[1:] + [len(block)]
    sample_locs = zip(sample_starts, sample_ends, strict=False)

    return [_parse_sample(block[x[0] : x[1]], haplotypes) for x in sample_locs]


def _parse_haplotypes(haplotype_block: str) -> pd.Series:
    """Parses haplotype block and returns pd.Series

    Parameters
    ----------
    haplotype_block: str
        String starting with `>> Haplotypes`

    Returns
    -------
    pd.Series
        Series named `sequence` with the number of haplotype as  an index
    """
    lines = [x.strip() for x in haplotype_block.split("\n") if x and not x.startswith(">")]
    idx = [int(x.split()[0]) for x in lines]
    seq = [x.split()[1] for x in lines]

    return pd.Series(data=seq, index=idx, name="sequence")


def _parse_sample(sample_block: str, haplotypes: pd.Series) -> pd.DataFrame:
    """Parse sample block and return a single haplotypexread dataframe for the prob. matrix

    Parameters
    ----------
    sample_block: str
        String of a sample block
    haplotypes: pd.Series
        haplotype sequence series

    Returns
    -------
    pd.DataFrame
    """
    reads_start = sample_block.index(">>> Reads")
    reads_end = sample_block.index(">>> Matrix")
    try:
        full_matrix_start = sample_block.index(">>> Read->Haplotype in Full")
    except ValueError:
        full_matrix_start = None

    reads = [x for x in sample_block[reads_start:reads_end].split("\n") if x and not x.startswith(">")]
    read_names = [x.strip().split()[1] for x in reads]
    if full_matrix_start is None:
        matrix = [x for x in sample_block[reads_end:].split("\n") if x and not x.startswith(">")]
    else:
        matrix = [x for x in sample_block[reads_end:full_matrix_start].split("\n") if x and not x.startswith(">")]
    array = np.vstack([np.fromstring(x, dtype=float, sep=" ") for x in matrix])
    result = pd.DataFrame(data=array, index=haplotypes.index, columns=read_names)
    result["sequence"] = haplotypes
    return result


def parse_haplotype_matrix_file(filename: str) -> dict:
    """Parses a file that contains haplotype matrices, return
    dictionary by location of haplotype x read matrix

    Parameters
    ----------
    filename: str
        File name

    Returns
    -------
    dict
    """
    with open(filename, encoding="latin-1") as fid:
        filecontents = fid.read()
    blocks = [x for x in re.split(r"> Location ", filecontents) if x]

    locations = [_extract_location(x) for x in blocks]
    block_dfs = [_parse_active_region(x) for x in blocks]
    return dict(zip(locations, block_dfs, strict=False))



import argparse
import sys

from simppl.simple_pipeline import SimplePipeline
from ugbio_core.filter_bed import intersect_bed_regions


def get_parser():
    parser = argparse.ArgumentParser(prog="intersect_bed_regions.py", description=run.__doc__)
    parser.add_argument(
        "--include-regions",
        nargs="+",
        required=True,
        help="List of paths to BED files to be intersected.",
    )
    parser.add_argument(
        "--exclude-regions",
        nargs="*",
        default=None,
        help="List of paths to BED or VCF files to be subtracted from the intersected result.",
    )
    parser.add_argument("--output-bed", default="output.bed", help="Path to the output BED file.")
    parser.add_argument(
        "--assume-input-sorted",
        action="store_true",
        help="If set, assume that the input files are already sorted. "
        "If not set, the function will sort them on-the-fly.",
    )
    parser.add_argument(
        "--max-mem",
        type=int,
        default=None,
        help="Maximum memory in bytes allocated for the sort-bed operations. If not specified, "
        "the function will allocate 80 percent of the available system memory.",
    )

    return parser


def run(argv):
    """Intersect BED regions with the option to subtract exclude regions."""
    parser = get_parser()
    SimplePipeline.add_parse_args(parser)
    args = parser.parse_args(argv[1:])
    sp = SimplePipeline(args.fc, args.lc, debug=args.d, print_timing=True)
    intersect_bed_regions(
        include_regions=args.include_regions,
        exclude_regions=args.exclude_regions,
        output_bed=args.output_bed,
        assume_input_sorted=args.assume_input_sorted,
        max_mem=args.max_mem,
        sp=sp,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



from matplotlib import pyplot as plt

SMALL_SIZE = 12
MEDIUM_SIZE = 18
BIGGER_SIZE = 26
TITLE_SIZE = 36
FIGSIZE = (16, 8)
GRID = True


def set_pyplot_defaults(
    title_size=TITLE_SIZE,
    small_size=SMALL_SIZE,
    medium_size=MEDIUM_SIZE,
    bigger_size=BIGGER_SIZE,
    grid=GRID,
    figsize=FIGSIZE,
):
    plt.rc("font", size=small_size)  # controls default text sizes
    plt.rc("axes", titlesize=title_size)  # fontsize of the axes title
    plt.rc("axes", labelsize=bigger_size)  # fontsize of the x and y labels
    plt.rc("axes", grid=grid)  # is grid on
    plt.rc("xtick", labelsize=medium_size)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=medium_size)  # fontsize of the tick labels
    plt.rc("legend", fontsize=medium_size)  # legend fontsize
    plt.rc("figure", titlesize=title_size)  # fontsize of the figure title
    plt.rc("figure", figsize=figsize)  # size of the figure



import json
import re

import h5py
import pandas as pd
from pandas import DataFrame
from ugbio_core.logger import logger


def get_h5_keys(h5_filename: str):
    with pd.HDFStore(h5_filename, "r") as store:
        keys = store.keys()
        return keys


def should_skip_h5_key(key: str, ignored_h5_key_substring: str):
    if ignored_h5_key_substring is None:
        return None
    return ignored_h5_key_substring in key


def preprocess_h5_key(key: str):
    result = key
    if result[0] == "/":
        result = result[1:]
    return result


def preprocess_columns(dataframe):
    """Handle multiIndex/ hierarchical .h5 - concatenate the columns for using it as single string in JSON."""

    def flatten_multi_index(col, separator):
        flat = separator.join(col)
        flat = re.sub(f"{separator}$", "", flat)
        return flat

    if hasattr(dataframe, "columns"):
        if isinstance(dataframe.columns, pd.core.indexes.multi.MultiIndex):
            dataframe.columns = [flatten_multi_index(col, "___") for col in dataframe.columns.to_numpy()]


def convert_h5_to_json(
    input_h5_filename: str,
    root_element: str,
    ignored_h5_key_substring: str = None,
    output_json: str = None,
):
    """Convert an .h5 metrics file to .json with control over the root element and the processing

    Parameters
    ----------
    input_h5_filename: str
        Input h5 file name

    root_element: str
        Root element of the returned json

    ignored_h5_key_substring: str, optional
        A way to filter some of the keys using substring match, if None (default) none are filtered

    output_json : str, optional
        Output json file name to create if not None (default)

    Returns
    -------
    str
        The result json string includes the schema (the types) of the metrics as well as the metrics themselves.

    """

    new_json_dict = {root_element: {}}
    h5_keys = get_h5_keys(input_h5_filename)
    for h5_key in h5_keys:
        if should_skip_h5_key(h5_key, ignored_h5_key_substring):
            logger.warning("Skipping: %s", h5_key)
            continue
        logger.info("Processing: %s", h5_key)
        data_frame = read_hdf(input_h5_filename, h5_key)
        preprocess_columns(data_frame)
        data_frame_to_json = data_frame.to_json(orient="table")
        json_dict = json.loads(data_frame_to_json)
        new_json_dict[root_element][preprocess_h5_key(h5_key)] = json_dict

    if output_json:
        with open(output_json, "w", encoding="utf-8") as outfile:
            json.dump(new_json_dict, outfile, indent=4)
    json_string = json.dumps(new_json_dict, indent=4)
    return json_string


def read_hdf(  # noqa: C901 #TODO: refactor. too complex
    file_name: str,
    key: str = "all",
    skip_keys: list[str] | None = None,
    columns_subset: list[str] | None = None,
) -> DataFrame:
    """
    Read data-frame or data-frames from an h5 file

    Parameters
    ----------
    file_name: str
        path of local file
    key: str
        hdf key to the data-frame.
        Special keys:
        1. all - read all data-frames from the file and concat them
        2. all_human_chrs - read chr1, ..., chr22, chrX keys, and concat them
        3. all_somatic_chrs - chr1, ..., chr22
    skip_keys: Iterable[str]
        collection of keys to skip from reading the H5 (e.g. concordance, input_args ... )
    columns_subset: list[str], optional
        select a subset of columns

    Returns
    -------
    data-frame or concat data-frame read from the h5 file according to key
    """
    if skip_keys is None:
        skip_keys = []
    if key == "all":
        with h5py.File(file_name, "r") as h5_file:
            keys = list(h5_file.keys())
        for k in skip_keys:
            if k in keys:
                keys.remove(k)
        dfs = []
        for k in keys:
            tmpdf: DataFrame = DataFrame(pd.read_hdf(file_name, key=k))
            if columns_subset is not None:
                tmpdf = tmpdf[[x for x in columns_subset if x in tmpdf.columns]]
            if tmpdf.shape[0] > 0:
                dfs.append(tmpdf)
        return pd.concat(dfs)
    if key == "all_human_chrs":
        dfs = [DataFrame(pd.read_hdf(file_name, key=f"chr{x}")) for x in list(range(1, 23)) + ["X", "Y"]]
        return pd.concat(dfs)
    if key == "all_hg19_human_chrs":
        dfs = [DataFrame(pd.read_hdf(file_name, key=x)) for x in list(range(1, 23)) + ["X", "Y"]]
        return pd.concat(dfs)
    if key == "all_somatic_chrs":
        dfs = [DataFrame(pd.read_hdf(file_name, key=f"chr{x}")) for x in list(range(1, 23))]
        return pd.concat(dfs)
    # If not one of the special keys:
    return DataFrame(pd.read_hdf(file_name, key=key))



from __future__ import annotations

import numpy as np
import pysam
from ugbio_core import array_utils as utils
from ugbio_core import dna_sequence_utils as dnautils
from ugbio_core.consts import DEFAULT_FLOW_ORDER

from . import read_expander

error_rates = np.logspace(0.001, 0.01, 10)

# We always allow for a bit of error
# _CONFUSION_MATRIX = np.load(pjoin(_workdir, "confusion.npy"))
# min_prob = np.min(_CONFUSION_MATRIX[_CONFUSION_MATRIX > 0])
# CONFUSION_MATRIX = np.clip(_CONFUSION_MATRIX, min_prob, None)
# CONFUSION_MATRIX = (CONFUSION_MATRIX.T / CONFUSION_MATRIX.sum(axis=1)).T
CONFUSION_MATRIX = np.nan


def generate_flow_matrix(gtr_calls: np.ndarray, confusion_matrix: np.ndarray = CONFUSION_MATRIX) -> np.ndarray:
    """
    generate_flow_matrixa flow matrix from the ground truth flows with low probabilities for

    Parameters
    ----------
    gtr_calls : np.ndarray
            Ground truth in key space
    confusion_matrix : np.ndarray
            n_hmer x n_hmer confusion matrix - rows - ground truth, columns - probabilities

    Returns
    -------
    modified_flow_matrix : np.ndarray
            Modified matrix corresponding to mutations in the read, in the mutated positions the
            probabilities will be around 0.5 and 0.5 between the ground truth and the mutation,
            in other positions - they will be
            according to the probabilities of the confusion matrix
    """

    expected_reads = confusion_matrix[gtr_calls, :]
    cum_expected = np.cumsum(expected_reads, axis=1)
    rng = np.random.default_rng()
    rands = rng.uniform(0, 1, cum_expected.shape[0])
    mutation_profile = utils.searchsorted2d(cum_expected, rands)

    # this is the mutation profile in each nucleotide
    diff = mutation_profile - gtr_calls

    # something narrow around 0 and 1
    mutations = np.nonzero(diff)[0]
    rng = np.random.default_rng()
    mutation_probs = rng.beta(3, 3, size=len(mutations))
    non_mutation_probs = 1 - mutation_probs

    flow_matrix = confusion_matrix[gtr_calls, :].T
    minconf = confusion_matrix.min()
    mutation_flow_matrix = np.ones((confusion_matrix.shape[0], len(mutations))) * minconf

    mutation_flow_matrix[gtr_calls[mutations], range(len(mutations))] = non_mutation_probs
    mutation_flow_matrix[mutation_profile[mutations], range(len(mutations))] = mutation_probs
    mutation_flow_matrix = mutation_flow_matrix / mutation_flow_matrix.sum(axis=0)
    flow_matrix[:, mutations] = mutation_flow_matrix
    return flow_matrix


def get_flow2_base(flow_order: str, desired_length: int) -> np.ndarray:
    """
    Given order of flows - what base corresponds to each flow

    Parameters
    ----------
    flow_order: str
            String of flow orders
    desired_length: int
            Needed length of flows
    Returns
    -------
    np.ndarray
            Array of bases
    """
    flow_order = np.tile(np.array(list(flow_order)), int(desired_length / len(flow_order)) + 1)
    flow_order = flow_order[:desired_length]
    return flow_order


def get_rflow_order(flow_matrix: np.ndarray, flow_order: str = DEFAULT_FLOW_ORDER) -> str:
    """Finds the flowget of the first base of the reverse of the expected sequence
    Parameters
    ----------
    flow_matrix: np.ndarray
        Flow matrix
    flow_order: str
        Flow order that generated the flow matrix

    Returns
    -------
    str
        Reversed flow order
    """
    tmp = flow_order * flow_matrix.shape[1]
    tmp = tmp[: flow_matrix.shape[1]]
    rflow = dnautils.revcomp(tmp)
    return rflow[: len(flow_order)]


def key_space2_base_space(flow_matrix: np.ndarray, flow_order: str = DEFAULT_FLOW_ORDER) -> tuple:
    """
    Converts from matrix of hmer probabilities to sequence (most probable)

    Parameters
    ----------
    flow_matrix: np.ndarray
            n_hmer x flow matrix of hmer probabilities
    flow_order : str
            flow order (default: TCAG)

    Returns
    -------
    tuple (str, np.ndarray, np.ndarray)
            1. String of most probable bases in the read
            2. Array of the last output base BEFORE each flow (For instance for flows 01021 the output is -1 -1 0 0 2 )
            3. Array of the call for each flow (integer array)

    """
    flow_order = get_flow2_base(flow_order, flow_matrix.shape[1])
    n_reps = np.argmax(flow_matrix, axis=0)
    n_reps_shifted = np.array([0] + list(np.argmax(flow_matrix, axis=0)))[:-1]
    flow2base = -1 + np.cumsum(n_reps_shifted)
    return "".join(np.repeat(flow_order, n_reps)), flow2base, n_reps


def identify_mutations(
    flow_matrix: np.ndarray,
    n_mutations: int,
    threshold: float,
    gtr_calls: np.ndarray | None = None,
) -> tuple:
    """
    Identifies mutations to report in the read

    Parameters
    ----------
    flow_matrix: np.ndarray
            Modified flow matrix with mutations
    n_mutations: int
            Maximal number of mutations to return
    threshold: float
            Minimal mutation probability
    gtr_calls: np.ndarray
            In case there is a ground truth - calls will be calculated relative to the ground truth

    Returns
    -------
    tuple:
            (np.ndarray, np.ndarray) List of flows in key space and list of key values corresponding to each mutation

    Note
    ----
    All mutations that have probability of n_mutation strongest mutations are passed (potentially more than n_mutations)
    """

    mutation_id_matrix = flow_matrix.copy()
    mutation_id_matrix[mutation_id_matrix < threshold] = 0

    if gtr_calls is None:
        gtr_calls = np.argmax(flow_matrix, axis=0)

    # do not identify ground truth calls
    mutation_id_matrix[gtr_calls, np.arange(flow_matrix.shape[1])] = 0

    strongest_candidates = np.argsort(mutation_id_matrix, axis=None)[::-1]

    # in case of several mutations with the same probability - pass all of them
    # this is needed to prevent cases of different reverse complement mutations
    cur_idx = n_mutations - 1
    while (
        mutation_id_matrix.flat[strongest_candidates[cur_idx]] > 0
        and mutation_id_matrix.flat[strongest_candidates[cur_idx]]
        == mutation_id_matrix.flat[strongest_candidates[n_mutations - 1]]
    ):
        cur_idx += 1
    n_mutations = cur_idx
    strongest_candidates = strongest_candidates[:n_mutations]
    # convert flat index that argsort returns to rows and columns
    take = mutation_id_matrix.flat[strongest_candidates] > 0
    tmp = np.unravel_index(strongest_candidates[take], mutation_id_matrix.shape)
    return tmp[1], tmp[0]


def get_likelihood(flow_matrix: np.ndarray, gtr_calls: np.ndarray | None = None) -> float:
    """
    Return log likelihood of the path in flow matrix

    Parameters
    ----------
    flow_matrix: np.ndarray
            n_hmers x n_flows matrix of hmer call probabilities
    gtr_calls: Optional[np.ndarray]
            None or specific path in flows to return likelihood. If none - optimal path is returned

    Returns
    -------
    Log2 of the likelihood of the flow
    """

    lflow = np.log2(flow_matrix)
    if gtr_calls is not None:
        likelihood = lflow[gtr_calls, np.arange(lflow.shape[1])].sum()
    else:
        likelihood = np.max(lflow, axis=0).sum()
    return likelihood


def get_mutation_change(
    gtr_calls: np.ndarray,
    mutation: tuple,
    base_calls: str,
    key2base: np.ndarray,
    flow_order: str = DEFAULT_FLOW_ORDER,
) -> tuple:
    """
    Return the sequence that the mutation changes

    Parameters
    ----------
    gtr_calls: np.ndarray
            List of original (unmutated hmers)
    mutation: tuple
            (int, int) - flow index, new hmer
    base_calls: str
            ground truth read
    key2base: np.ndarray
            For each flow - what was the last generated base BEFORE the call
    flow_order: str
            Flow order

    Returns
    -------
    tuple : (int, str, str)
            Pos, old sequence, new sequence
    """

    flow_order = "".join(get_flow2_base(flow_order, len(gtr_calls)))

    ref_alleles = gtr_calls[mutation[0]]
    alt_alleles = mutation[1]
    diff = ref_alleles - alt_alleles
    if diff == 0:
        raise ValueError("No mutation in this position")
    seq_base = key2base[mutation[0]]
    if diff > 0:
        # DELETION
        if seq_base >= 0:
            start_pos = seq_base
            start_seq = base_calls[start_pos : start_pos + diff + 1]
            end_seq = base_calls[start_pos]
        elif seq_base < 0:
            # If we are deleting the first bases in the key
            start_pos = 0
            start_seq = base_calls[start_pos : start_pos + diff + 1]
            end_seq = base_calls[start_pos + diff]

    elif diff < 0:
        # INSERTION
        start_pos = seq_base
        if seq_base >= 0:
            start_seq = base_calls[start_pos : start_pos + 1]
            end_seq = base_calls[start_pos : start_pos + 1] + flow_order[mutation[0]] * (-diff)
        elif seq_base == -1:
            start_pos = 0
            start_seq = base_calls[start_pos : start_pos + 1]
            end_seq = flow_order[mutation[0]] * (-diff) + base_calls[start_pos : start_pos + 1]
    if start_pos < 0:
        raise ValueError("Assertion failed - start_pos < 0")
    if len(start_seq) == len(end_seq):
        raise ValueError("Assertion failed - no difference b/w alleles")
    return start_pos, start_seq, end_seq


def get_log_lik_diff(flow_matrix: np.ndarray, gtr_calls: np.ndarray, mutation: tuple):
    """
    Reurns difference in log likelihood of reads

    Parameters
    ----------
    flow_matrix: np.ndarray
            n_hmer x n_flows matrix
    gtr_calls: np.ndarray
            Calls in key space
    mutation: tuple
            int - number of flow, int - new hmer
    """
    lflow = np.log2(flow_matrix)
    original = lflow[gtr_calls[mutation[0]], mutation[0]]
    new = lflow[mutation[1], mutation[0]]
    return new - original


def seq_to_record(
    seq: str,
    rname: str,
    llk: float,
    alts: list = None,
    ralts: list = None,
) -> pysam.AlignedSegment:
    """Convert string to pysam record

    Parameters
    ----------
    seq: str
        Read sequence
    rname: str
        Read name
    llk: Optional[float]
        Log likelihood of the sequence
    alts: list
        List of tuples (pos, ref, alt, diff)
    ralts: list
        Same as alts, but for reverse complement sequence
    Returns
    -------
    pysam.AlignedSegment
         unaligned BAM record
    """

    seg = pysam.AlignedSegment()

    seg.query_sequence = seq
    seg.is_unmapped = True
    seg.query_name = rname
    seg.query_qualities = pysam.qualitystring_to_array("I" * len(seq))
    seg.reference_name = "*"
    seg.next_reference_name = "*"
    seg.reference_start = 0
    seg.mapping_quality = 255
    seg.cigarstring = "*"

    def alt2str(allele_repr):
        return "%d,%s,%s,%.2f" % allele_repr

    alts = sorted(alts, key=lambda x: (x[3], x[0]))
    # note different sorting order for ralts so that it is the same as alts when we reverse
    ralts = sorted(ralts, key=lambda x: (x[3], -x[0]))
    rseq = dnautils.revcomp(seq)

    for alt in alts:
        if not validate_alt(seq, alt):
            raise ValueError("Weird mutation sequence")
    for ralt in ralts:
        if not validate_alt(rseq, ralt):
            raise ValueError("Weird rmutation sequence")
    valts = [validate_rcalt_v2(*x, seq, rseq) for x in zip(alts, ralts, strict=False)]
    for valt in valts:
        if not valt:
            raise ValueError("Weird rvariant")
    alts = sorted(alts, key=lambda x: x[0])
    ralts = sorted(ralts, key=lambda x: (x[0]))

    alts = ";".join([alt2str(x) for x in alts])
    ralts = ";".join([alt2str(x) for x in ralts])
    seg.tags = [("LL", llk), ("AL", alts), ("RA", ralts)]

    return seg


def validate_rcalt_v2(alt: tuple, ralt: tuple, seq: str, rseq: str) -> bool:
    """Check that the read after application of ralt is the same as after applying alt

    Parameters
    ----------
    alt: tuple
        Alternative allele
    ralt: tuple
        Revcomp
    seq: str
        Sequence
    rseq: str
        Revcomp
    """
    if apply_variant(rseq, ralt) == dnautils.revcomp(apply_variant(seq, alt)):
        return True
    return False


def apply_variant(seq: str, variant: tuple) -> str:
    """Apply variant to sequence"""
    return read_expander.get_read_variant_str(seq, [variant])


def validate_alt(seq: str, alt: tuple) -> bool:
    """Check that the reference alleles are right

    Parameters
    ----------
    seq: str
        Sequence of the read
    alt: tuple
        Mutation

    Returns
    bool
    """
    pos = alt[0]
    if pos < 0 or pos >= len(seq):
        return False
    if seq[pos : pos + len(alt[1])] == alt[1]:
        return True
    return False



from __future__ import annotations

import collections
import itertools
import multiprocessing
import os
import pickle
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd
import pyBigWig as pbw  # noqa: N813
import pyfaidx
import pysam
import ugbio_core.dna_sequence_utils as dnautils
from simppl.simple_pipeline import SimplePipeline
from ugbio_core.consts import (
    CYCLE_SKIP,
    CYCLE_SKIP_DTYPE,
    CYCLE_SKIP_STATUS,
    IS_CYCLE_SKIP,
    NON_CYCLE_SKIP,
    POSSIBLE_CYCLE_SKIP,
    UNDETERMINED_CYCLE_SKIP,
)
from ugbio_core.exec_utils import print_and_execute
from ugbio_core.flow_format import flow_based_read
from ugbio_core.logger import logger
from ugbio_core.vcfbed import bed_writer

UNDETERMINED = "NA"


class VcfAnnotator(ABC):
    """
    An abstract base class for annotating VCF files.
    """

    @abstractmethod
    def __init__(self):
        """
        Initializer of the base class.
        Derived classes should override this method if they require additional parameters.
        """

    @abstractmethod
    def edit_vcf_header(self, header: pysam.VariantHeader) -> pysam.VariantHeader:
        """
        Method to edit VCF header, accepts a pysam header, modifies and returns it.
        Derived classes should override this method to provide specific functionality.

        Parameters
        ----------
        header : pysam.VariantHeader
            VCF file header to be edited.

        Returns
        -------
        pysam.VariantHeader
            Edited VCF file header.
        """

    @abstractmethod
    def process_records(self, records: list[pysam.VariantRecord]) -> list[pysam.VariantRecord]:
        """
        Method to process VCF records. Accepts a list of pysam VariantRecords, modifies each and returns the list
        Derived classes should override this method to provide specific functionality.

        Parameters
        ----------
        records : list[pysam.VariantRecord]
            List of VCF records to be processed.

        Returns
        -------
        list[pysam.VariantRecord]
            Processed VCF records.
        """

    @staticmethod
    def merge_temp_files(
        contig_output_vcfs: list[str], output_path: str, header: pysam.VariantHeader = None, process_number: int = 1
    ):
        """
        Static method to merge temporary output files and write to the final output file.

        Args:
        contigs : list[str]
            List of contig names.
        output_path : str
            Path to the final output file.
        header : pysam.VariantHeader
            VCF file header.
        """
        try:
            # merge using bcftools
            command = (
                f"bcftools concat --threads {process_number} -Oz --naive -o {output_path} "
                f"{' '.join(contig_output_vcfs)}"
            )
            print_and_execute(command, shell=True)  # noqa: S604

            if header is not None:
                # reheader using bcftools
                command = f"bcftools reheader --threads {process_number} -h {header} {output_path} -o {output_path}"
                print_and_execute(command, shell=True)  # noqa: S604

            # index the final output file
            command = f"bcftools index --threads {process_number} -f -t {output_path}"
            print_and_execute(command, shell=True)  # noqa: S604

        finally:
            # remove the temporary files
            for temp_file_path in contig_output_vcfs:
                if os.path.exists(temp_file_path):
                    os.remove(temp_file_path)

    @staticmethod
    def process_vcf(
        annotators: list[VcfAnnotator],
        input_path: str,
        output_path: str,
        chunk_size: int = 10000,
        process_number: int = 1,
    ):
        """
        Static method to process a VCF file in chunks with multiple VcfAnnotator objects.
        Optionally runs in parallel over different contigs.

        Parameters
        ----------
        annotators : list[VcfAnnotator]
            List of VcfAnnotator objects.
        input_path : str
            Path to the input VCF file.
        output_path : str
            Path to the output file.
        chunk_size : int, optional
            The chunk size. Defaults to 10000.
        process_number : int, optional
            Number of processes to use. If N < 1, use all-available - abs(N) cores. Defaults to 1.
        """
        logger.info(f"Starting VCF annotation with {len(annotators)} annotators")
        # pickle the annotators
        logger.info(f"Pickling annotators to {output_path}.annotators.pickle")
        annotators_pickle = output_path + ".annotators.pickle"
        with open(annotators_pickle, "wb") as f:
            pickle.dump(annotators, f)

        out_dir = os.path.dirname(output_path)
        # Open the input VCF file
        logger.info(f"Reading input VCF file {input_path}")
        with pysam.VariantFile(input_path) as input_variant_file:
            logger.info("Reading and editing input VCF header")
            new_header = input_variant_file.header
            for annotator in annotators:
                new_header = annotator.edit_vcf_header(new_header)

            # Get the contigs
            logger.info("Getting contigs")
            contigs = list(input_variant_file.header.contigs)
            tmp_output_paths = []
            sp = SimplePipeline(0, max(100, len(contigs)))
            commands = []
            variant_counts = {}  # Dictionary to store variant counts per contig
            skipped_contigs = []
            for contig in contigs:
                try:
                    next(input_variant_file.fetch(contig))  # to raise StopIteration if contig is empty
                    variant_count = 1
                    # Iterate through variants in the contig and count them
                    # for _ in input_variant_file.fetch(contig):
                    #     variant_count += 1

                    if variant_count > 0:
                        out_per_contig = os.path.join(out_dir, contig + ".vcf.gz")
                        commands.append(
                            f"annotate_contig --vcf_in {input_path} --vcf_out {out_per_contig} "
                            f"--annotators_pickle {annotators_pickle} --contig {contig} --chunk_size {chunk_size}"
                        )
                        tmp_output_paths.append(out_per_contig)
                        variant_counts[contig] = variant_count
                    else:
                        logger.info(f"Skipping empty contig {contig}. No variants found.")
                except StopIteration:  # as e:
                    skipped_contigs.append(contig)
                    # logger.info(f"Skipping empty contig {contig}. Error: {e}")
            # for contig, count in variant_counts.items():
            #     if count>0:
            #         logger.info(f"Contig {contig} has {count} variants.")
            if len(skipped_contigs) > 0:
                logger.info(f"Number of skipped contigs: {len(skipped_contigs)}")
                logger.info(f"First 10 skipped contigs: {skipped_contigs[:10]}")
            # process_number<1 it is interpreted as all the CPUs except -process_number
            if process_number < 1:
                process_number = multiprocessing.cpu_count() + process_number

            if process_number > 1:  # multiprocessing
                sp.run_parallel(commands, process_number)
            else:
                for command in commands:
                    sp.print_and_run(command)
            # Merge the temporary output files into the final output file and remove them
            VcfAnnotator.merge_temp_files(tmp_output_paths, output_path, process_number=process_number)

    @staticmethod
    def process_contig(
        vcf_in: str,
        vcf_out: str,
        annotators: list[VcfAnnotator],
        contig: str,
        chunk_size: int,
    ):
        """
        Static helper method to process a single contig in chunks.
        Args:
        vcf_in : str
            The input VCF file.
        annotators : list[VcfAnnotator]
            List of VcfAnnotator objects.
        contig : str
            The contig to process.
        chunk_size : int
            The chunk size.
        output_path : str
            Path to the output file.
        """

        with pysam.VariantFile(vcf_in) as input_variant_file:
            # Edit the header - needed in case new INFO/FORMATS fields are added
            new_header = input_variant_file.header
            for annotator in annotators:
                new_header = annotator.edit_vcf_header(new_header)
            # Write output file
            with pysam.VariantFile(vcf_out, "w", header=new_header) as output_variant_file:
                records = []
                for record in input_variant_file.fetch(contig):
                    records.append(record)

                    if len(records) == chunk_size:
                        for annotator in annotators:
                            records = annotator.process_records(records)

                        for rec in records:
                            output_variant_file.write(rec)

                        records = []

                # Process the remaining records
                if records:
                    for annotator in annotators:
                        records = annotator.process_records(records)

                    for record in records:
                        output_variant_file.write(record)
        pysam.tabix_index(vcf_out, preset="vcf", force=True)


def classify_indel(concordance: pd.DataFrame) -> pd.DataFrame:
    """Classifies indel as insertion or deletion

    Parameters
    ----------
    concordance: pd.DataFrame
        Dataframe of concordances

    Returns
    -------
    pd.DataFrame:
        Modifies dataframe by adding columns "indel_classify" and "indel_length"
    """

    def classify(x):
        if not x["indel"]:
            return None
        if len(x["ref"]) < max(len(y) for y in x["alleles"]):
            return "ins"
        return "del"

    concordance["indel_classify"] = concordance.apply(classify, axis=1, result_type="reduce")
    concordance["indel_length"] = concordance.apply(
        lambda x: max(abs(len(y) - len(x["ref"])) for y in x["alleles"]),
        axis=1,
        result_type="reduce",
    )
    return concordance


def is_hmer_indel(concordance: pd.DataFrame, fasta_file: str) -> pd.DataFrame:
    """Checks if the indel is hmer indel and outputs its length
    Note: The length of the shorter allele is output.

    Parameters
    ----------
    concordance: pd.DataFrame
        Dataframe of concordance or of VariantCalls. Should contain a collumn "indel_classify"
    fasta_file: str
        FAI indexed fasta file

    Returns
    -------
    pd.DataFrame
        Adds column hmer_indel_length, hmer_indel_nuc
    """

    fasta_idx = pyfaidx.Fasta(fasta_file, build_index=False, rebuild=False)

    def _is_hmer(rec, fasta_idx):
        if not rec["indel"]:
            return (0, None)

        if rec["indel_classify"] == "ins":
            alt = [x for x in rec["alleles"] if x != rec["ref"]][0][1:]
            if len(set(alt)) == 1 and fasta_idx[rec["chrom"]][rec["pos"]].seq.upper() == alt[0]:
                return (
                    dnautils.hmer_length(fasta_idx[rec["chrom"]], rec["pos"]),
                    alt[0],
                )
        elif rec["indel_classify"] == "del":
            del_seq = rec["ref"][1:]
            if (
                len(set(del_seq)) == 1
                and fasta_idx[rec["chrom"]][rec["pos"] + len(rec["ref"]) - 1].seq.upper() == del_seq[0]
            ):
                return (
                    len(del_seq) + dnautils.hmer_length(fasta_idx[rec["chrom"]], rec["pos"] + len(rec["ref"]) - 1),
                    del_seq[0],
                )
        return (0, None)

    results = concordance.apply(lambda x: _is_hmer(x, fasta_idx), axis=1, result_type="reduce")
    concordance["hmer_indel_length"] = [x[0] for x in results]
    concordance["hmer_indel_nuc"] = [x[1] for x in results]
    return concordance


def get_motif_around_snv(record: pysam.VariantRecord, size: int, faidx: pyfaidx.Fasta):
    """ "
    Extract sequence around an SNV, of size "size" on each side with the central base being the ref base

    Parameters
    ----------
    record: pysam.VariantRecord
        Record, must be an SNV or result will be wrong
    size: int
        Size of motif on each side
    faidx: pyfaidx.Fasta
        Fasta index

    Returns
    -------
    """
    if not isinstance(record, pysam.VariantRecord):
        raise TypeError(f"record must be pysam.VariantRecord, got {type(record)}")
    chrom = faidx[record.chrom]
    pos = record.pos
    size = int(min(size, pos - 1, len(chrom) - pos))
    if size <= 0:
        raise ValueError(f"size must be positive, got {size}")

    return chrom[pos - size - 1 : pos + size].seq.upper()


def get_motif_around_snp(record: pysam.VariantRecord, size: int, faidx: pyfaidx.Fasta):
    """
    Extract sequence around an SNV, of size "size" on each side with the central base being the ref base

    Parameters
    ----------
    record: pysam.VariantRecord
        PySam Record, must be an SNV
    size: int
        Size of motif on each side
    faidx: pyfaidx.Fasta
        Fasta index

    Returns
    -------
    tuple
        (left_motif, right_motif) - left and right motifs

    """
    # make sure the input is a vcf record of an SNV
    if not isinstance(record, pysam.VariantRecord):
        raise TypeError(f"record must be pysam.VariantRecord, got {type(record)}")
    if not (len(record.ref) == 1 and len(record.alts) >= 1 and len(record.alts[0]) == 1):
        raise ValueError("Not an SNV")
    # make sure the size is positive
    size = int(size)
    if size <= 0:
        raise ValueError(f"size must be positive, got {size}")
    # make sure the fasta index is valid
    if not isinstance(faidx, pyfaidx.Fasta):
        raise TypeError(f"faidx must be pyfaidx.Fasta, got {type(faidx)}")
    # extract the sequence
    chrom = faidx[record.chrom]
    pos = record.pos
    return (
        chrom[pos - size - 1 : pos - 1].seq.upper(),
        chrom[pos : pos + size].seq.upper(),
    )


def get_motif_around(concordance: pd.DataFrame, motif_size: int, fasta: str) -> pd.DataFrame:
    """Extract sequence around the indel

    Parameters
    ----------
    concordance: pd.DataFrame
        Concordance dataframe
    motif_size: int
        Size of motif on each side
    fasta: str
        Indexed fasta

    Returns
    -------
    pd.DataFrame
        DataFrame. Adds "left_motif" and right_motif
    """

    def _get_motif_around_snp(rec, size, faidx):
        chrom = faidx[rec["chrom"]]
        pos = rec["pos"]
        return (
            chrom[pos - size - 1 : pos - 1].seq.upper(),
            chrom[pos : pos + size].seq.upper(),
        )

    def _get_motif_around_non_hmer_indel(rec, size, faidx):
        chrom = faidx[rec["chrom"]]
        pos = rec["pos"]
        return (
            chrom[pos - size : pos].seq.upper(),
            chrom[pos + len(rec["ref"]) - 1 : pos + len(rec["ref"]) - 1 + size].seq.upper(),
        )

    def _get_motif_around_hmer_indel(rec, size, faidx):
        chrom = faidx[rec["chrom"]]
        pos = rec["pos"]
        hmer_length = rec["hmer_indel_length"]
        return (
            chrom[pos - size : pos].seq.upper(),
            chrom[pos + hmer_length : pos + hmer_length + size].seq.upper(),
        )

    def _get_motif(rec, size, faidx):
        if rec["indel"] and rec["hmer_indel_length"] > 0:
            return _get_motif_around_hmer_indel(rec, size, faidx)
        if rec["indel"] and rec["hmer_indel_length"] == 0:
            return _get_motif_around_non_hmer_indel(rec, size, faidx)
        return _get_motif_around_snp(rec, size, faidx)

    faidx = pyfaidx.Fasta(fasta, build_index=False, rebuild=False)
    tmp = concordance.apply(lambda x: _get_motif(x, motif_size, faidx), axis=1, result_type="reduce")
    concordance["left_motif"] = [x[0] for x in list(tmp)]
    concordance["right_motif"] = [x[1] for x in list(tmp)]
    return concordance


def get_gc_content(concordance: pd.DataFrame, window_size: int, fasta: str) -> pd.DataFrame:
    """Extract sequence around the indel

    Parameters
    ----------
    concordance: pd.DataFrame
        Concordance dataframe
    window_size: int
        Size of window for GC calculation (around start pos of variant)
    fasta: str
        Indexed fasta

    Returns
    -------
    pd.DataFrame
        DataFrame. Adds "left_motif" and right_motif
    """

    def _get_gc(rec, size, faidx):
        chrom = faidx[rec["chrom"]]
        beg = rec["pos"] - int(size / 2)
        end = beg + size
        seq = chrom[beg:end].seq.upper()
        seq_gc = seq.replace("A", "").replace("T", "")
        return float(len(seq_gc)) / len(seq)

    faidx = pyfaidx.Fasta(fasta, build_index=False, rebuild=False)
    tmp = concordance.apply(lambda x: _get_gc(x, window_size, faidx), axis=1, result_type="reduce")
    concordance["gc_content"] = list(tmp)
    return concordance


def get_coverage(
    df: pd.DataFrame,
    bw_coverage_files_high_quality: list[str],
    bw_coverage_files_low_quality: list[str],
) -> pd.DataFrame:
    """Adds coverage columns to the variant dataframe. Three columns are added: coverage - total coverage,
    well_mapped_coverage - coverage of reads with mapping quality > min_quality and repetitive_read_coverage -
    which is the difference between the two.
    bw_coverage_files should be outputs of `coverage_analysis.py` and have .chr??. inside the name

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe (VCF or concordance)
    bw_coverage_files_high_quality : List[str]
        List of BW for the coverage at the high MAPQ quality threshold
    bw_coverage_files_low_quality : List[str]
        List of BW for the coverage at the MAPQ threshold 0


    Returns
    -------
    pd.DataFrame
        Modified dataframe
    """

    chrom2f = [(list(pbw.open(f).chroms().keys())[0], f) for f in bw_coverage_files_high_quality]
    chrom2bw_dict = dict(chrom2f)

    chrom2f = [(list(pbw.open(f).chroms().keys())[0], f) for f in bw_coverage_files_low_quality]
    for v_var in chrom2f:
        chrom2bw_dict[v_var[0]] = (chrom2bw_dict[v_var[0]], v_var[1])

    df.insert(len(df.columns), "coverage", np.nan)
    df.insert(len(df.columns), "well_mapped_coverage", np.nan)
    df.insert(len(df.columns), "repetitive_read_coverage", np.nan)

    gdf = df.groupby("chrom")

    for g_var in gdf.groups:
        if g_var in chrom2bw_dict:
            bw_hq = chrom2bw_dict[g_var][0]
            bw_lq = chrom2bw_dict[g_var][1]
        else:
            continue

        with pbw.open(bw_hq) as bw_file:
            values_hq = [bw_file.values(g_var, x[1] - 1, x[1])[0] for x in gdf.groups[g_var]]
        with pbw.open(bw_lq) as bw_file:
            values_lq = [bw_file.values(g_var, x[1] - 1, x[1])[0] for x in gdf.groups[g_var]]
        df.loc[gdf.groups[g_var], "coverage"] = values_lq
        df.loc[gdf.groups[g_var], "well_mapped_coverage"] = values_hq
        df.loc[gdf.groups[g_var], "repetitive_read_coverage"] = np.array(values_lq) - np.array(values_hq)
    return df


def annotate_intervals(df: pd.DataFrame, annotfile: str) -> tuple[pd.DataFrame, str]:
    """
    Adds column based on interval annotation file (T/F)

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe to be annotated
    annotfile: str
        bed file of the annotated intervals

    Returns
    -------
    pd.DataFrame
        Adds boolean column for the annotation (parsed from the file name)
    str
        Annotation name


    """
    annot = annotfile.split("/")[-1]
    if annot[-4:] == ".bed":
        annot = annot[:-4]

    df[annot] = False
    annot_df = bed_writer.parse_intervals_file(annotfile)
    gdf = df.groupby("chrom")
    gannot_df = annot_df.groupby("chromosome")
    for chrom in gdf.groups.keys():
        if chrom not in gannot_df.groups:
            continue
        gdf_ix = gdf.groups[chrom]
        gannot_ix = gannot_df.groups[chrom]
        pos1 = np.array(df.loc[gdf_ix, "pos"])
        pos2 = np.array(annot_df.loc[gannot_ix, "start"])
        pos1_closest_pos2_start = np.searchsorted(pos2, pos1) - 1
        pos2 = np.array(annot_df.loc[gannot_ix, "end"])
        pos1_closest_pos2_end = np.searchsorted(pos2, pos1)

        is_inside = pos1_closest_pos2_start == pos1_closest_pos2_end
        df.loc[gdf_ix, annot] = is_inside
    return df, annot


def fill_filter_column(df: pd.DataFrame) -> pd.DataFrame:
    """Fills filter status column with PASS when there are missing values
    (e.g. when the FILTER column in the vcf has dots, or when false negative variants
    were added to the dataframe)

    Parameters
    ----------
    df : pd.DataFrame
        Description

    Returns
    -------
    pd.DataFrame
        Description
    """
    if "filter" not in df.columns:
        df["filter"] = np.nan
    fill_column_locs = (pd.isna(df["filter"])) | (df["filter"] == "")
    df.loc[fill_column_locs, "filter"] = "PASS"
    return df


# TODO: Refactor. too complex
def annotate_cycle_skip(df: pd.DataFrame, flow_order: str, gt_field: str = None) -> pd.DataFrame:  # noqa: C901
    """Adds cycle skip information: non-skip, NA, cycle-skip, possible cycle-skip

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe
    flow_order : str
        Flow order
    gt_field: str
        If snps that correspond to a specific genotype are to be considered

    Returns
    -------
    pd.DataFrame
        Dataframe with columns "cycleskip_status" - possible values are:
        * non-skip (not a cycle skip at any flow order)
        * NA (undefined - non-snps, multiallelic snps )
        * cycle-skip
        ( possible cycle-skip (cycle skip at a different flow order))
    """

    def is_multiallelic(x, gt_field):
        s_var = set(x[gt_field]) | {0}
        if len(s_var) > 2:  # noqa: PLR2004
            return True
        return False

    def get_non_ref(x):
        return [y for y in x if y != 0][0]

    def is_non_polymorphic(x, gt_field):
        s_var = set(x[gt_field]) | {0}
        return len(s_var) == 1

    if gt_field is None:
        na_pos = df["indel"] | (df["alleles"].apply(len) > 2)  # noqa: PLR2004
    else:
        na_pos = (
            df["indel"]
            | df.apply(lambda x: is_multiallelic(x, gt_field), axis=1)
            | df.apply(lambda x: is_non_polymorphic(x, gt_field), axis=1)
        )
    df["cycleskip_status"] = UNDETERMINED
    snp_pos = ~na_pos
    snps = df.loc[snp_pos].copy()
    left_last = np.array(snps["left_motif"]).astype(np.bytes_)
    right_first = np.array(snps["right_motif"]).astype(np.bytes_)

    ref = np.array(snps["ref"]).astype(np.bytes_)
    if gt_field is None:
        alt = np.array(snps["alleles"].apply(lambda x: x[1] if len(x) > 1 else None)).astype(np.bytes_)
    else:
        nra = snps[gt_field].apply(get_non_ref)
        snps["nra_idx"] = nra
        snps.loc[snps.nra_idx.isna(), "nra_idx"] = 1
        snps["nra_idx"] = snps["nra_idx"].astype(int)
        alt = np.array(snps.apply(lambda x: x["alleles"][x["nra_idx"]], axis=1)).astype(np.bytes_)
        snps = snps.drop("nra_idx", axis=1)

    ref_seqs = np.char.add(np.char.add(left_last, ref), right_first)
    alt_seqs = np.char.add(np.char.add(left_last, alt), right_first)

    ref_encs = [
        _catch(
            flow_based_read.generate_key_from_sequence,
            str(np.char.decode(x)),
            flow_order,
            exception_type=ValueError,
            handle=lambda x: UNDETERMINED,
        )
        for x in ref_seqs
    ]
    alt_encs = [
        _catch(
            flow_based_read.generate_key_from_sequence,
            str(np.char.decode(x)),
            flow_order,
            exception_type=ValueError,
            handle=lambda x: UNDETERMINED,
        )
        for x in alt_seqs
    ]

    cycleskip = np.array(
        [
            x
            for x in range(len(ref_encs))
            if isinstance(ref_encs[x], np.ndarray)
            and isinstance(alt_encs[x], np.ndarray)
            and len(ref_encs[x]) != len(alt_encs[x])
        ]
    )
    poss_cycleskip = [
        x
        for x in range(len(ref_encs))
        if isinstance(ref_encs[x], np.ndarray)
        and isinstance(alt_encs[x], np.ndarray)
        and len(ref_encs[x]) == len(alt_encs[x])
        and (
            np.any(ref_encs[x][ref_encs[x] - alt_encs[x] != 0] == 0)
            or np.any(alt_encs[x][ref_encs[x] - alt_encs[x] != 0] == 0)
        )
    ]

    undetermined = np.array(
        [
            x
            for x in range(len(ref_encs))
            if (isinstance(ref_encs[x], str) and ref_encs[x] == UNDETERMINED)
            or (isinstance(alt_encs[x], str) and alt_encs[x] == UNDETERMINED)
        ]
    )
    s_var = set(np.concatenate((cycleskip, poss_cycleskip, undetermined)))
    non_cycleskip = [x for x in range(len(ref_encs)) if x not in s_var]

    vals = [""] * len(snps)
    for x in cycleskip:
        vals[x] = "cycle-skip"
    for x in poss_cycleskip:
        vals[x] = "possible-cycle-skip"
    for x in non_cycleskip:
        vals[x] = "non-skip"
    for x in undetermined:
        vals[x] = UNDETERMINED
    snps["cycleskip_status"] = vals

    df.loc[snp_pos, "cycleskip_status"] = snps["cycleskip_status"]
    return df


def get_cycle_skip_dataframe(flow_order: str):
    """
    Generate a dataframe with the cycle skip status of all possible SNPs. The resulting dataframe is 192 rows, with the
    multi-index "ref_motif" and "alt_motif", which are the ref and alt of a SNP with 1 flanking base on each side
    (trinucleotide context). This output can be readily joined with a vcf dataframe once the right columns are created.

    Parameters
    ----------
    flow_order

    Returns
    -------

    """
    # build index composed of all possible SNPs
    ind = pd.MultiIndex.from_tuples(
        [
            x
            for x in itertools.product(
                ["".join(x) for x in itertools.product(["A", "C", "G", "T"], repeat=3)],
                ["A", "C", "G", "T"],
            )
            if x[0][1] != x[1]
        ],
        names=["ref_motif", "alt_motif"],
    )
    df_cskp = pd.DataFrame(index=ind).reset_index()
    df_cskp.loc[:, "alt_motif"] = (
        df_cskp["ref_motif"].str.slice(0, 1) + df_cskp["alt_motif"] + df_cskp["ref_motif"].str.slice(-1)
    )
    df_cskp.loc[:, CYCLE_SKIP_STATUS] = df_cskp.apply(
        lambda row: determine_cycle_skip_status(row["ref_motif"], row["alt_motif"], flow_order),
        axis=1,
    ).astype(CYCLE_SKIP_DTYPE)
    df_cskp.loc[:, IS_CYCLE_SKIP] = df_cskp[CYCLE_SKIP_STATUS] == CYCLE_SKIP
    return df_cskp.set_index(["ref_motif", "alt_motif"]).sort_index()


def determine_cycle_skip_status(ref: str, alt: str, flow_order: str):
    """return the cycle skip status, expects input of ref and alt sequences composed of 3 bases where only the 2nd base
    differs"""
    if len(ref) != 3 or len(alt) != 3 or ref[0] != alt[0] or ref[2] != alt[2] or ref == alt:  # noqa: PLR2004
        raise ValueError(
            f"""Invalid inputs ref={ref}, alt={alt}
expecting input of ref and alt sequences composed of 3 bases where only the 2nd base differs"""
        )
    try:
        ref_key = np.trim_zeros(flow_based_read.generate_key_from_sequence(ref, flow_order), "f")
        alt_key = np.trim_zeros(flow_based_read.generate_key_from_sequence(alt, flow_order), "f")
        if len(ref_key) != len(alt_key):
            return CYCLE_SKIP

        for r_val, a_val in zip(ref_key, alt_key, strict=False):
            if (r_val != a_val) and ((r_val == 0) or (a_val == 0)):
                return POSSIBLE_CYCLE_SKIP
        return NON_CYCLE_SKIP
    except ValueError:
        return UNDETERMINED_CYCLE_SKIP


def get_trinuc_sub_list():
    trinuc_list = ["".join([a, b, c]) for a in "ACGT" for b in "ACGT" for c in "ACGT"]
    trinuc_sub = [
        ">".join([a, b]) for a in trinuc_list for b in trinuc_list if (a[0] == b[0]) & (a[2] == b[2]) & (a[1] != b[1])
    ]
    return trinuc_sub


def parse_trinuc_sub(rec: pysam.VariantRecord, faidx: pyfaidx.Fasta = None):
    """
    Parse the trinucleotide substitution from a pysam VariantRecord

    Parameters
    ----------
    rec: pysam.VariantRecord
        pysam VariantRecord of an SNV
    faidx: pyfaidx.Fasta, optional
        pyfaidx.Fasta object

    Returns
    -------
    str
        trinucleotide substitution in the format "ref>alt" or None if the motif contains an unknown base

    Raises
    ------
    AssertionError
        If rec is not a pysam VariantRecord of an SNV
    AssertionError
        If faidx is not a pyfaidx.Fasta object
    ValueError
        If faidx is None and X_LM and X_RM are not in the VCF
    """
    # check inputs
    if not isinstance(rec, pysam.VariantRecord):
        raise TypeError(f"rec must be pysam.VariantRecord, got {type(rec)}")
    if not (len(rec.ref) == 1 and len(rec.alts) == 1 and len(rec.alts[0]) == 1):
        raise ValueError("Not an SNV")
    if faidx is not None:
        if not isinstance(faidx, pyfaidx.Fasta):
            raise TypeError(f"faidx must be pyfaidx.Fasta, got {type(faidx)}")
    # get left and right bases in the reference
    if "X_LM" in rec.info and "X_RM" in rec.info:
        # Pre-annotated mode: X_LM and X_RM are the left and right motif sequences, respectively
        # obtained from AnnotateVcf.wdl pipeline
        left_motif = rec.info["X_LM"][0][-1]
        right_motif = rec.info["X_RM"][0][0]
    else:
        # Non-pre-annotated mode: get the left and right bases from the fasta index
        if faidx is None:
            raise ValueError("faidx must be provided if X_LM and X_RM are not in the VCF")
        left_motif, right_motif = get_motif_around_snp(rec, 1, faidx)
    motif_ref = left_motif + rec.ref + right_motif
    motif_alt = left_motif + rec.alts[0] + right_motif
    # if motif has an unknown base, return None
    if "N" in motif_ref or "N" in motif_alt:
        return None
    return motif_ref + ">" + motif_alt


def get_trinuc_substitution_dist(vcf_file, ref_fasta: str = None):
    """
    Get the SNV trinucleotide substitution distribution from a VCF file

    Parameters
    ----------
    vcf_file: str
        Path to VCF file
    ref_fasta: str, optional
        Path to reference fasta file

    Returns
    -------
    dict
        Dictionary of trinucleotide substitutions and their counts

    Raises
    ------
    AssertionError
        If vcf_file is not a string
    AssertionError
        If vcf_file is not a valid path
    AssertionError
        If ref_fasta is not a string
    AssertionError
        If ref_fasta is not a valid path
    """
    # check inputs
    if not isinstance(vcf_file, str):
        raise TypeError(f"vcf_file must be string, got {type(vcf_file)}")
    if not os.path.exists(vcf_file):
        raise FileNotFoundError(f"vcf_file {vcf_file} does not exist")
    if ref_fasta is not None:
        if not isinstance(ref_fasta, str):
            raise TypeError(f"ref_fasta must be string, got {type(ref_fasta)}")
        if not os.path.exists(ref_fasta):
            raise FileNotFoundError(f"ref_fasta {ref_fasta} does not exist")
    # initialize trinuc_sub dictionary
    trinuc_sub = get_trinuc_sub_list()
    trinuc_dict = {k: 0 for k in trinuc_sub}
    # initialize fasta index if given
    faidx = pyfaidx.Fasta(ref_fasta, build_index=False, rebuild=False) if ref_fasta is not None else None
    with pysam.VariantFile(vcf_file, "rb") as infh:
        for rec in infh.fetch():
            # check if rec is a biallelic SNV
            if len(rec.ref) == 1 and len(rec.alts) == 1 and len(rec.alts[0]) == 1:
                trinuc_sub = parse_trinuc_sub(rec, faidx)
                if trinuc_sub is not None:
                    trinuc_dict[trinuc_sub] += 1
    return trinuc_dict


def _catch(
    func: collections.abc.Callable,
    *args,
    exception_type: Exception = Exception,
    handle: collections.abc.Callable = lambda e: e,
    **kwargs,
):
    """From https://stackoverflow.com/a/8915613 general wrapper that
    catches exception and returns value. Useful for handling error in list comprehension

    Parameters
    ----------
    func: Callable
        Function being wrapped
    *args:
        func non-named arguments
    exception_type:
        Type of exception to catch
    handle:
        How to handle the exception (what value to return)
    **kwargs:
        func named arguments
    """
    try:
        return func(*args, **kwargs)
    except exception_type as e:
        return handle(e)



"""pileuptofreq.py - Convert samtools pileup output to frequency table
Copyright (C) 2022
Author: Valentin Maurer <valentin.maurer@stud.uni-heidelberg.de>
"""

import fileinput
from argparse import ArgumentParser
from collections import defaultdict
from pathlib import Path

import pandas as pd

PILEUP_IGNORE = {",", ".", "*", "#", ">", "<"}
USE_STARTEND = True
# BASE_TO_FORWAWRD = {
#     "A" : "A", "C" : "C", "T" : "T", "N" : "N",
#     "a" : "T", "c" : "G", "t" : "A", "n" : "N"
# }


def pileup_to_freq(reference: str, pileup: str) -> int:  # noqa: C901 #TODO: refactor. too complex
    """Counts A, C, G, T, In, Del occurence given a samtools pileup string"""

    pileup = pileup.upper()
    frequencies = defaultdict(int)
    is_indel = False
    indel_pos = 0
    indel_size = ""

    is_start = 0
    last_base = ""

    for k in pileup:
        if k in ("+", "-"):
            is_indel = True
            indel_pos = 0
            indel_size = ""
            indel = k
            continue

        if k == "$":
            frequencies[last_base] -= not USE_STARTEND
            continue

        # Handle first base in read; Indels would come after (^)(QUAL)(BASE) structure
        if k == "^":
            is_start = 1
            continue
        if is_start > 0:
            if is_start < 2:  # noqa: PLR2004
                is_start += 1
            elif is_start == 2:  # noqa: PLR2004
                frequencies[k] += USE_STARTEND
                is_start = 0
            continue

        if is_indel:
            is_digit = str.isdigit(k)
            if is_digit:
                indel_size += k
                continue
            indel_pos += 1
            indel += k
            if indel_pos == int(indel_size):
                frequencies[indel] += 1
                is_indel = False
                indel_pos = 0
            continue

        last_base = k
        frequencies[k] += 1

    frequencies[reference] += frequencies[","] + frequencies["."]
    frequencies["Deletion"] = frequencies["*"] + frequencies["#"]

    return {k: v for k, v in frequencies.items() if v > 0 and k not in PILEUP_IGNORE}


def create_frequncies_from_pileup(input_pileup_file) -> pd.DataFrame:
    """Create a pandas DataFrame summerizing samtools pileup file results"""
    df_frequencies = pd.DataFrame(columns=["Depth", "Chrom", "Pos", "Ref", "Sample", "Base", "Count"])
    data = []
    with fileinput.input(input_pileup_file, mode="r") as fin:
        for line in fin:
            sample = 0
            values = line.split("\t")
            chr_name = values[0]
            pos = values[1]
            ref = values[2]
            dp = values[3]
            # POSITION  = ';'.join([str(values[i]) for i in range(3)])
            # Go through all seq fields
            for i in range(4, len(values), 3):
                sample += 1
                # No reads aligned to this position
                if values[i - 1] == "0":
                    continue
                freq = pileup_to_freq(values[2], values[i])
                for base, count in freq.items():
                    data.append([dp, chr_name, pos, ref, sample, base, count])
    fin.close()
    df_frequencies = pd.DataFrame(data, columns=["Depth", "Chrom", "Pos", "Ref", "Sample", "Base", "Count"])
    return df_frequencies


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "-i", type=Path, help="Samtools mpileup output without extra flags. If empty, stdin is used", default=None
    )
    parser.add_argument("-o", type=Path, help="Output file, if empty stdout is used.", default=None)
    args = parser.parse_args()

    if args.i is None:
        args.i = ("-",)

    df_frequencies = create_frequncies_from_pileup(args.i)
    df_frequencies.to_csv(args.o, sep=";", index=False)



#!/env/python
import argparse
import sys

from ugbio_core import h5_utils


def run(argv: list[str]):
    ap = argparse.ArgumentParser(
        prog="convert_h5_to_json.py",
        description="Convert h5 file to json file (using pandas table serialization) optionally ready for MongoDB",
    )
    ap.add_argument("--input_h5", help="Metrics input h5 file name", required=True, type=str)
    ap.add_argument("--output_json", help="Metrics output json file name", required=True, type=str)
    ap.add_argument("--root_element", help="Root element of the output JSON", required=True, type=str)
    ap.add_argument(
        "--ignored_h5_key_substring",
        help="The h5 key substring to ignore during conversion",
        required=False,
        type=str,
    )

    args = ap.parse_args(argv[1:])
    input_h5_filename = args.input_h5
    output_json_filename = args.output_json
    ignored_h5_key_substring = args.ignored_h5_key_substring
    root_element = args.root_element

    json_string = h5_utils.convert_h5_to_json(input_h5_filename, root_element, ignored_h5_key_substring)
    with open(output_json_filename, "w", encoding="utf-8") as f:
        f.write(json_string)


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



from enum import Enum

from pandas.api.types import CategoricalDtype


class FileExtension(Enum):
    """File Extension enum"""

    PARQUET = ".parquet"
    HDF = ".hdf"
    H5 = ".h5"
    CSV = ".csv"
    TSV = ".tsv"
    BAM = ".bam"
    CRAM = ".cram"
    PNG = ".png"
    JPEG = ".jpeg"
    FASTA = ".fasta"
    TXT = ".txt"
    VCF = ".vcf"
    VCF_GZ = ".vcf.gz"


DEFAULT_FLOW_ORDER = "TGCA"
CHROM_DTYPE = CategoricalDtype(
    categories=[f"chr{j}" for j in range(1, 23)] + ["chrX", "chrY", "chrM"],
    ordered=True,
)

# cycle skip
CYCLE_SKIP_STATUS = "cycle_skip_status"
CYCLE_SKIP = "cycle-skip"
POSSIBLE_CYCLE_SKIP = "possible-cycle-skip"
NON_CYCLE_SKIP = "non-skip"
UNDETERMINED_CYCLE_SKIP = "NA"
IS_CYCLE_SKIP = "is_cycle_skip"
CYCLE_SKIP_DTYPE = CategoricalDtype(categories=[CYCLE_SKIP, POSSIBLE_CYCLE_SKIP, NON_CYCLE_SKIP], ordered=True)

# vcf
CHROM = "chrom"
POS = "pos"
REF = "ref"
ALT = "alt"
QUAL = "qual"
FILTER = "filter"
GT = "GT"
AD = "AD"
DP = "DP"
VAF = "VAF"



from __future__ import annotations

import itertools
import os
import re
import subprocess
import sys
import tempfile
from collections import Counter

import numpy as np
import pandas as pd
import pysam
import tqdm

ERROR_PROBS = "probability.csv"


def get_matrix(testmatrices: np.ndarray, idx: int) -> np.ndarray:
    """Summary

    Parameters
    ----------
    testmatrices : np.ndarray
        Description
    idx : int
        Description

    Returns
    -------
    np.ndarray
        Description
    """
    if len(testmatrices.shape) == 4:  # noqa: PLR2004
        return testmatrices[idx, 0, :, :].T
    return testmatrices[idx, :, :].T


def save_tmp_kr_matrix(tensor_name: str, n_classes: int, n_flows: int, output_dir: str) -> str:
    """extract kr npy matrix from the probability tensor sequence

    Parameters
    ----------
    tensor_name : str
        File with a probability tensor
    n_classes : int
        number of classes (in case tensort sequence is not npy)
    n_flows : int
        number of flows (in case tensort sequence is not npy)
    output_dir: str
        Location to save the kr file

    Returns
    -------
    Temporary file with Kr
    """

    if not tensor_name.endswith("npy"):
        testmatrices = np.memmap(tensor_name, dtype=np.float32)
        testmatrices = testmatrices.reshape(-1, 1, n_flows, n_classes)
    else:
        testmatrices = np.load(tensor_name, mmap_mode="r")
    kr_tag = np.squeeze(np.argmax(testmatrices, axis=3))
    _, tempname = tempfile.mkstemp(suffix=".npy", dir=output_dir)
    np.save(tempname, kr_tag)
    return tempname


def get_kr(key_matrix: np.ndarray, idx: int, scale_factor: int = 100) -> np.ndarray:
    """Returns a row of the key matrix

    Parameters
    ----------
    key_matrix : np.ndarray
        flowx x max_hmer array
    idx : int
        row index to return
    scale_factor : int, optional
        scale factor of the result

    Returns
    -------
    np.ndarray
        Row of key matrix divided by the scale factor
    """
    return (key_matrix[idx, :] + scale_factor // 2) // scale_factor


def key2base(key: np.ndarray, flow_order: str) -> str:
    """Convert key to base sequence

    Parameters
    ----------
    key : np.ndarray
        Flow key
    flow_order : str
        Flow order

    Returns
    -------
    str
        Sequence
    """
    return "".join([k * n for k, n in zip(flow_order, key, strict=False)])


def _calculate_correct_prob(matrix: np.ndarray) -> np.ndarray:
    """Calculates the probability to be correct call at each flow
    The probabilities of errors other than the second highest are
    discarded

    Parameters
    ----------
    matrix: np.ndarray
        Flow matrix
    """
    kr_tag = np.argmax(matrix, axis=0)
    high_err = np.argpartition(matrix, -2, axis=0)[-2, :]
    return matrix[kr_tag, np.arange(matrix.shape[1])] / (
        matrix[kr_tag, np.arange(matrix.shape[1])] + matrix[high_err, np.arange(matrix.shape[1])]
    )


def _generate_quality(seq: str, kr_tag: np.ndarray = None, correct_prob: np.ndarray = None) -> str:
    """Summary

    Parameters
    ----------
    seq : str
        Sequence
    kr_tag : np.ndarray
        Kr (optional)
    correct_prob: np.ndarray
        probability (optional)

    Returns
    -------
    str
        I of the length of the sequence
    """

    if kr_tag is None or correct_prob is None:
        return "I" * len(seq)

    error_prob = 1 - correct_prob
    probs = error_prob[kr_tag > 0]
    tmp_kr = kr_tag[kr_tag > 0]
    ends = np.cumsum(tmp_kr) - 1
    starts = ends + 1 - tmp_kr
    probs = probs / 2
    output = np.zeros(tmp_kr.sum())
    output[ends] = output[ends] + probs
    output[starts] = output[starts] + probs
    output = np.clip(output, 0.0001, None)
    error_phred = (-10 * np.log10(output)).astype(np.int8) + 33
    quals = error_phred.view("c").tostring().decode()
    return quals


def write_sequences(tensor_name: str, seq_file_name: str, n_flows: int, n_classes: int, flow_order: str) -> int:
    """Write convert the prob. tensor to sequences and write them to the file

    Parameters
    ----------
    tensor_name : str
        Name of the probability tensor
    seq_file_name : str
        Name of the file to write the sequences
    n_flows : int
        Number of flows
    n_classes : int
        Number of classes
    flow_order: str
        Flow cycle

    Returns
    ------------------
    int
        Number of records written
    """
    if not tensor_name.endswith("npy"):
        testmatrices = np.memmap(tensor_name, dtype=np.float32)
        testmatrices = testmatrices.reshape(-1, 1, n_flows, n_classes)
    else:
        testmatrices = np.load(tensor_name, mmap_mode="r")

    flow_order = (flow_order * n_flows)[:n_flows]
    count = 0
    with open(seq_file_name, "w", encoding="latin-1") as out:
        for i in range(testmatrices.shape[0]):
            matrix = get_matrix(testmatrices, i)

            kr_tag = np.argmax(matrix, axis=0)
            seq = key2base(kr_tag, flow_order)
            correct_prob = _calculate_correct_prob(matrix)
            qual = _generate_quality(seq, kr_tag, correct_prob)
            if len(seq) != len(qual):
                raise ValueError("Sequence of a different length with quality")
            out.write(f"{seq}\t{qual}\n")
            count += 1
    return count


def extract_tags(bamfile: str, output_file: str, tag_list: list) -> None:
    """Extracts specified tags and writes them into the TSV file

    Parameters
     ----------
    bamfile : str
         Input BAM file
    output_file : str
         Output txt file
    tag_list : list
         List of tags to extract
    """
    with open(output_file, "w", encoding="latin-1") as out:
        with pysam.AlignmentFile(bamfile, check_sq=False) as inp:
            for rec in inp:
                tags = [rec.get_tag(x, with_value_type=True) for x in tag_list]
                out.write("\t".join([f"{tag_list[i]}:{tags[i][1]}:{tags[i][0]}" for i in range(len(tags))]))
                out.write("\n")


def matrix_to_sparse(
    matrix: np.ndarray,
    kr_tag: np.ndarray,
    probability_threshold: float = 0,
    probability_sf: float = 10,
) -> tuple:
    """Summary

    Parameters
    ----------
    matrix : np.ndarray
        Input matrix
    kr_tag : np.ndarray
        regressed key
    probability_threshold : float, optional
        threshold **ratio** to report
    probability_sf: float
        Phred scaling factor (default - 10)
    Returns
    -------
    tuple
        Row, column, probability ratio normalized to the kr
    """
    max_hmer = matrix.shape[0] - 1
    probability_threshold = -probability_sf * np.log10(probability_threshold)

    tmp_matrix = matrix.copy()
    kr_clip = np.clip(kr_tag, 0, max_hmer)
    kr_val = tmp_matrix[kr_clip, np.arange(len(kr_tag))]
    tmp_matrix[kr_clip, np.arange(len(kr_tag))] = 0

    row, column = np.nonzero(tmp_matrix)
    values = tmp_matrix[row, column]

    values = np.log10(values)
    norm_value = np.log10(np.clip(kr_val[column], 1e-10, None))
    normalized_values = -probability_sf * (values - norm_value)
    normalized_values = np.clip(normalized_values, -6 * probability_sf, 6 * probability_sf).astype(np.int16)
    suppress = normalized_values > probability_threshold
    return row[~suppress], column[~suppress], normalized_values[~suppress]


def array_repr(input_array) -> str:
    return ",".join([str(y) for y in input_array])


def write_matrix_tags(
    tensor_name: str,
    key_name: str,
    output_file: str,
    n_flows: int = 280,
    n_classes: int = 13,
    probability_threshold: float = 0.003,
    probability_sf: float = 10,
) -> tuple[int, int]:
    """Writes probability tensor into the text file

    Parameters
    ----------
    tensor_name : str
        Name of the tensor file
    key_name : str
        Regressed key file name
    output_file : str
        Name of the output file
    n_flows : int, optional
        Number of flows (default: 280)
    n_classes : int, optional
        Number of classes called (default: 13)
    probability_threshold : float, optional
        Minimal probability to report
    probability_sf: float, optional
        Scaling factor for phred probability


    Returns
    -------
    int
        Number of sequences written
    """

    if key_name is not None:
        if not key_name.endswith("npy"):
            key = np.memmap(key_name, dtype=np.int16).reshape((-1, n_flows))
        else:
            key = np.load(key_name, mmap_mode="r")

    else:
        key = None
    if not tensor_name.endswith("npy"):
        testmatrices = np.memmap(tensor_name, dtype=np.float32)
        testmatrices = testmatrices.reshape(-1, 1, n_flows, n_classes)
    else:
        testmatrices = np.load(tensor_name, mmap_mode="r")
        if len(testmatrices.shape) == 3:  # noqa: PLR2004
            testmatrices = testmatrices[:, np.newaxis, ...]

    print(f"Read {testmatrices.shape[0]} predictions", flush=True, file=sys.stderr)
    empty = 0
    complete = 0
    with open(output_file, "w", encoding="latin-1") as out:
        for idx in tqdm.tqdm(range(testmatrices.shape[0])):
            matrix = get_matrix(testmatrices, idx)
            if key is not None:
                kr_tag = key[idx, :]
            else:
                kr_tag = np.argmax(matrix, axis=0)

            kh_tag, kf_tag, kd_tag = matrix_to_sparse(matrix, kr_tag, probability_threshold, probability_sf)
            if len(kh_tag) == 0 or len(kf_tag) == 0 or len(kd_tag) == 0:
                out.write("\n")
                empty += 1
                continue
            kr_str = array_repr(kr_tag)
            kr_str = "kr:B:C," + kr_str
            kh_str = array_repr(kh_tag)
            kh_str = "kh:B:C," + kh_str
            kf_str = array_repr(kf_tag)
            kf_str = "kf:B:S," + kf_str
            kd_str = array_repr(kd_tag)
            kd_str = "kd:B:s," + kd_str
            out.write("\t".join((kr_str, kd_str, kh_str, kf_str)))
            out.write("\n")
            complete += 1
    return empty, complete


def extract_header(input_bam: str, output_sam: str) -> None:
    """Get SAM header

    Parameters
    ----------
    input_bam : str
        Input BAM file
    output_sam : str
        Output SAM file

    No Longer Returned
    ------------------
    None
    """
    with open(output_sam, "w", encoding="latin-1") as outfile:
        subprocess.check_call(["samtools", "view", "-H", input_bam], stdout=outfile)  # noqa: S607


def add_matrix_to_bam(
    input_bam: str,
    input_matrix: str,
    output_bam: str,
    replace_sequence_file: str = None,
) -> None:
    """Summary

    Parameters
    ----------
    input_bam : str
        Description
    input_matrix : str
        Description
    output_bam : str
        Description
    replace_sequence_file : str, optional
        Description
    """

    if replace_sequence_file is not None:
        rgbi_fname = output_bam + "rgbi.txt"
        extract_tags(input_bam, rgbi_fname, ["RG", "bi", "rq"])

    re_pipe, we_pipe = os.pipe()
    extract_header(input_bam, output_bam + ".hdr")
    process1 = subprocess.Popen(["cat", output_bam + ".hdr"], stdout=we_pipe)  # noqa: S607

    process4 = subprocess.Popen(["samtools", "view", "-b", "-o", output_bam, "-"], stdin=re_pipe)  # noqa: S607
    process1.wait()
    if replace_sequence_file is None:
        process2 = subprocess.Popen(["samtools", "view", input_bam], stdout=subprocess.PIPE)  # noqa: S607
    else:
        process2a = subprocess.Popen(["samtools", "view", input_bam], stdout=subprocess.PIPE)  # noqa: S607
        process2 = subprocess.Popen(["cut", "-f1-9"], stdin=process2a.stdout, stdout=subprocess.PIPE)  # noqa: S607
        process2a.stdout.close()

    if replace_sequence_file is None:
        subprocess.Popen(["paste", "-", input_matrix], stdin=process2.stdout, stdout=we_pipe)  # noqa: S607
    else:
        subprocess.Popen(
            ["paste", "-"] + [replace_sequence_file, rgbi_fname, input_matrix],
            stdin=process2.stdout,
            stdout=we_pipe,
        )

    process2.stdout.close()
    os.close(we_pipe)

    process4.wait()
    os.unlink(output_bam + ".hdr")
    if replace_sequence_file is not None:
        os.unlink(rgbi_fname)


def read_range_bytes(obj, start, end):
    """Summary

    Parameters
    ----------
    obj : TYPE
        Description
    start : TYPE
        Description
    end : TYPE
        Description

    Returns
    -------
    TYPE
        Description
    """
    range_header = "bytes=%d-%d" % (start, end - 1)
    tmp = obj.get(Range=range_header)["Body"].read()
    return np.frombuffer(tmp, dtype=np.int16)


def read_key_range(obj, start_idx, end_idx, read_size=280):
    """Summary

    Parameters
    ----------
    obj : TYPE
        Description
    start_idx : TYPE
        Description
    end_idx : TYPE
        Description
    read_size : int, optional
        Description

    Returns
    -------
    TYPE
        Description
    """
    start = start_idx * 2 * read_size
    end = end_idx * 2 * read_size
    tmp = read_range_bytes(obj, start, end)
    return tmp.reshape(-1, read_size)


def idx_to_block(idx, block_size_in_reads):
    """Summary

    Parameters
    ----------
    idx : TYPE
        Description
    block_size_in_reads : TYPE
        Description

    Returns
    -------
    TYPE
        Description
    """
    return idx // block_size_in_reads


def read_in_block(idx, block_size_in_reads):
    """Summary

    Parameters
    ----------
    idx : TYPE
        Description
    block_size_in_reads : TYPE
        Description

    Returns
    -------
    TYPE
        Description
    """
    return idx % block_size_in_reads


def block_idx_to_block_start_end(
    block_idx: int, block_size_in_reads: int, total_number_of_reads: int = 1150319344
) -> tuple:
    """Helper function that returns offset of start and end of the block in number of reads

    Parameters
    ----------
    block_idx : int
        Index of the block
    block_size_in_reads : int
        reads in block
    total_number_of_reads : int, optional
        maximal number of reads

    Returns
    -------
    tuple
        index of the first and the last read in the block
    """
    return (
        block_idx * block_size_in_reads,
        min(total_number_of_reads, (block_idx + 1) * block_size_in_reads),
    )


# TODO: no one imports this methid. consider removing
def read_error_probs(
    error_probs_csv: str = ERROR_PROBS,
    left_motif_size: int = 5,
    right_motif_size: int = 5,
    n_regression_bins: int = 0,
    *,
    binned_by_quality: bool = False,
) -> pd.DataFrame | list:
    """Read error probs CSV and produces data frame

    Parameters
    ----------
    error_probs_csv : str, optional
        CSV file with error probabilities. Example %s
    binned_by_quality : bool, optional
        Optional - True if the probabilities are binned by the regressed signal aof the base
    left_motif_size : int, optional
        Length of the left motif.
    right_motif_size : int, optional
        Length of the right motif
    n_regression_bins : int, optional
        Default - 0 - number of regression bins

    Returns
    -------
    Union[pd.DataFrame, list]
    """

    source_dataframe = pd.DataFrame(pd.read_csv(error_probs_csv))
    source_dataframe["left"] = source_dataframe["motif"].apply(lambda x: x[:left_motif_size])
    source_dataframe["right"] = source_dataframe["motif"].apply(lambda x: x[-right_motif_size:])
    source_dataframe["middle"] = source_dataframe["motif"].apply(
        lambda x: x[left_motif_size + 1 : -right_motif_size - 1]
    )
    source_dataframe["hmer_letter"] = source_dataframe["middle"].apply(lambda x: x[1])
    source_dataframe["hmer_number"] = source_dataframe["middle"].apply(lambda x: x[0]).astype(int)
    source_dataframe = source_dataframe.drop(["motif", "middle"], axis=1)
    tups = [tuple(x) for x in source_dataframe[["left", "hmer_number", "hmer_letter", "right"]].to_numpy()]
    source_dataframe.index = pd.MultiIndex.from_tuples(tups, names=["left", "hmer_number", "hmer_letter", "right"])

    source_dataframe = source_dataframe.drop(["left", "right", "hmer_letter", "hmer_number"], axis=1)

    if not ((n_regression_bins > 0 and binned_by_quality) or (n_regression_bins == 0 and not binned_by_quality)):
        raise ValueError("If not binned by quality - 0 bins")

    if binned_by_quality:
        n_diffs = int(source_dataframe.shape[1] / n_regression_bins)
        column_names = [f"{x-int((n_diffs-1)/2)}_{y}" for x in range(n_diffs) for y in range(n_regression_bins)]
        source_dataframe.columns = column_names
        return source_dataframe
    source_dataframe.columns = ["n(-1)", "n(0)", "n(+1)", "P(-1)", "P(0)", "P(+1)"]
    return source_dataframe


def _convert_to_probs(source_dataframe: pd.DataFrame):  # TODO: can be deteled
    """Converts counts to probabilities

    Parameters
    ----------
    source_dataframe : pd.DataFrame
            DataFrame of counts

    Returns
    -------
    adds columns P(-1), P(0), P(+1) to convert counts (n(-1), n(0),n(+1)) to probabilities

    """
    count_columns = [x for x in source_dataframe.columns if not x.startswith("P")]
    source_dataframe = source_dataframe[count_columns]
    prob_columns = [f"P({x})" for x in count_columns]
    sum_counts = source_dataframe[count_columns].sum(axis=1)
    sum_counts[sum_counts == 0] = 0.01
    probs = source_dataframe[count_columns].multiply(1 / sum_counts, axis=0)

    probs.columns = prob_columns
    source_dataframe[prob_columns] = probs
    return source_dataframe


def marginalize_error_probs(source_dataframe: pd.DataFrame, left_drop: int = 0, right_drop: int = 0) -> pd.DataFrame:
    """Marginalize error probabilities by combining motifs sharing common suffix (left) or prefix (right)
    This function is useful for calculation of error probabilities of nucleotides that are close to the end
    of the read

    Parameters
    ----------
    source_dataframe : pd.DataFrame
            Input DataFrame
    left_drop : int, optional
            Number of nucleotides to marginalize on in prefix of left context
    right_drop : int, optional
            Number of nucleotides to marginalize on in suffix of right context

    Returns
    -------
    pd.DataFrame
        Description
    """
    # source_dataframe = source_dataframe.drop(0, axis=0, level='hmer_number').copy()
    if not (left_drop > 0 or right_drop > 0):
        raise ValueError("No marginalization needed for these drop values")
    len_left = len(source_dataframe.index.get_level_values("left")[0])
    len_right = len(source_dataframe.index.get_level_values("right")[0])
    if not (left_drop <= len_left and right_drop <= len_right):
        raise ValueError("Unable to marginalize on more nucs than exist")

    groupby_left = source_dataframe.index.get_level_values("left").str[left_drop:]
    if right_drop > 0:
        groupby_right = source_dataframe.index.get_level_values("right").str[:-right_drop]
    else:
        groupby_right = source_dataframe.index.get_level_values("right")
    groupby_length = source_dataframe.index.get_level_values("hmer_number")
    groupby_hmer = source_dataframe.index.get_level_values("hmer_letter")

    renamed_index = pd.MultiIndex.from_arrays((groupby_left, groupby_right, groupby_length, groupby_hmer))
    source_dataframe1 = source_dataframe.copy()
    source_dataframe1.index = renamed_index
    gsource_dataframe = source_dataframe1.groupby(axis=0, level=[0, 1, 2, 3])
    source_dataframe1 = gsource_dataframe.agg(np.sum)
    # source_dataframe1 = _convert_to_probs(source_dataframe1)
    source_dataframe = source_dataframe1.reorder_levels(["left", "hmer_number", "hmer_letter", "right"], axis=0)
    # No need to add zero model since
    # source_dataframe = add_zero_model(source_dataframe)
    source_dataframe = source_dataframe.sort_index()
    return source_dataframe


def create_marginalize_dictionary(source_dataframe: pd.DataFrame) -> dict:
    """Creates a dictionary of all possible marginalizations of the error model

    Parameters
    ----------
    source_dataframe : pd.DataFrame
        original dataframe

    Returns
    -------
    dict
        dictionary with keys - sizes of the left and right motif

    Raises
    ------
    Exception
        Description
    """

    marginalize_dict = {}
    len_left = len(source_dataframe.index.get_level_values("left")[0])
    len_right = len(source_dataframe.index.get_level_values("right")[0])

    marginalize_dict[(len_left, len_right)] = source_dataframe.copy()
    for i in range(len_left + 1):
        for j in range(len_right + 1):
            if (len_left - i, len_right - j) in marginalize_dict:
                continue
            if (len_left - i + 1, len_right - j) in marginalize_dict:
                source_dataframe = marginalize_dict[(len_left - i + 1, len_right - j)]
                source_dataframe1 = marginalize_error_probs(source_dataframe, 1, 0)
                marginalize_dict[(len_left - i, len_right - j)] = source_dataframe1
            elif (len_left - i, len_right - j + 1) in marginalize_dict:
                source_dataframe = marginalize_dict[(len_left - i, len_right - j + 1)]
                source_dataframe1 = marginalize_error_probs(source_dataframe, 0, 1)
                marginalize_dict[(len_left - i, len_right - j)] = source_dataframe1
            else:
                raise Exception(f"Can't create {(len_left - i, len_right - j)}")
    return marginalize_dict


def add_zero_model(source_dataframe: pd.DataFrame) -> pd.DataFrame:
    """Add probabilities for 0->0 and 0->1 errors.
    Due to implementation difficulties P(0|0) and P(0|1) were not reported.
    The calculation is P(0|1) = P (1|0), P(0|0) = 1-P(1|0)

    Parameters
    ----------
    source_dataframe : pd.DataFrame
        Input dataframe

    Return
    ------
    pd.DataFrame

    Note
    ----
    This function does not work yet with the new format of the error model
    """
    new_source_dataframe = source_dataframe.xs(1, axis=0, level="hmer_number").copy()
    new_source_dataframe = pd.concat((new_source_dataframe,), keys=[0], names=["hmer_number"])
    new_source_dataframe = new_source_dataframe.reorder_levels(["left", "hmer_number", "hmer_letter", "right"], axis=0)

    # there is no meaning for counts for now in this case
    new_source_dataframe[["n(0)", "n(-1)", "n(+1)"]] = 0
    new_source_dataframe[["P(+1)"]] = new_source_dataframe[["P(-1)"]]
    new_source_dataframe[["P(-1)"]] = 0
    new_source_dataframe[["P(0)"]] = 1 - new_source_dataframe[["P(+1)"]]
    return pd.concat((new_source_dataframe, source_dataframe))


def split_by_signal_bins(source_dataframe: pd.DataFrame) -> pd.DataFrame:
    """Recieves dataframe with hmers binned by regression signal and splits it
    into a list by the bin of regression signal

    Parameters
    ----------
    source_dataframe : pd.DataFrame
        Source dataframe. Column names should be of the form ?(n_nn) where n is the difference
        between the read hmer and the ground truth hmer and nn is the value of the bin

    Returns
    -------
    pd.DataFrame
    """
    column_name_regexp = r"([A-Za-z]*)\({0,1}([\-0-9]+)_([\-0-9]+)\){0,1}"
    parsed_column_names = [re.search(column_name_regexp, x).groups() for x in source_dataframe.columns]
    parsed_column_names = sorted(parsed_column_names, key=(lambda x: (int(x[2]), x[0], int(x[1]))))
    dfs = []
    grouped_column_names = itertools.groupby(parsed_column_names, lambda x: (x[2]))
    for grouped_columns in grouped_column_names:
        take = list(grouped_columns[1])
        take_names = [x[0] + "(" * (len(x[0]) > 0) + x[1] + "_" + x[2] + ")" * (len(x[0]) > 0) for x in take]
        new_name = [x[0] + "(" * (len(x[0]) > 0) + x[1] + ")" * (len(x[0]) > 0) for x in take]
        new_name = [int(x) if x.isdigit() or x.startswith("-") else x for x in new_name]
        split_df = source_dataframe[take_names]

        split_df.columns = new_name
        dfs.append(split_df.copy())
    return dfs


def convert2_read_given_data(source_dataframe: pd.DataFrame) -> pd.DataFrame:
    """Convert probabilities in P(read | data = i ) to P(read=i | data)

    Parameters
    ----------
    source_dataframe : pd.DataFrame
        source dataframe. Should contain columns with counts with integer names (-2,-1,0,1,2 etc.)
        and probability columns (P(-2), P(-1)...) that represent e.g P(R=-1|H=0) (i.e. probability of
        deletion in read given haplotype for each row)

    Returns
    -------
    pd.DataFrame
        In each row now P(-1) now represents P(R=0 | H=-1) etc.
    """

    left_names = source_dataframe.index.get_level_values("left").unique()
    hmer_names = source_dataframe.index.get_level_values("hmer_letter").unique()
    right_names = source_dataframe.index.get_level_values("right").unique()
    hmer_number = source_dataframe.index.get_level_values("hmer_number").max()
    hmer_number = np.arange(0, hmer_number + 2)

    bins_number = int((source_dataframe.shape[1]) / 2)

    idx = pd.MultiIndex.from_product(
        [left_names, hmer_number, hmer_names, right_names],
        names=["left", "hmer_number", "hmer_letter", "right"],
    ).sort_values()
    result_dataframe = pd.DataFrame(index=idx, columns=source_dataframe.columns)

    dest_columns = []
    for err_idx in source_dataframe.columns[:bins_number]:
        source_dataframe[f"dest({err_idx})"] = (
            source_dataframe.index.get_level_values("hmer_number") - err_idx
        ).astype(int)
        dest_columns.append(f"dest({err_idx})")

    # now rearranging
    for err_idx in source_dataframe.columns[:bins_number]:
        tmp = source_dataframe[source_dataframe[f"dest({err_idx})"] >= 0]
        tmp = tmp.reset_index()
        tmp = tmp.sort_values(["left", f"dest({err_idx})", "hmer_letter", "right"])
        dest_index = pd.MultiIndex.from_frame(
            tmp[["left", f"dest({err_idx})", "hmer_letter", "right"]],
            names=["left", "hmer_number", "hmer_letter", "right"],
        )

        destination_contained = dest_index.isin(result_dataframe.index)
        dest_index = dest_index[destination_contained]
        result_dataframe.loc[dest_index, err_idx] = np.array(tmp[-err_idx].values)[destination_contained]
        result_dataframe.loc[dest_index, f"P({err_idx})"] = np.array(tmp[f"P({-err_idx})"].values)[
            destination_contained
        ]

    source_dataframe = source_dataframe.drop(dest_columns, axis=1)
    result_dataframe = result_dataframe.dropna(how="all")

    return result_dataframe


class ErrorModel:
    """Contains error model and functions to access it. The design of the class
    is mosty for efficiency purposes

    Attributes
    ----------
    error_model : TYPE
        Description
    hashed_dict : TYPE
        Description
    _em - error model. Currently implemented as numpy array, keeps only probabilities
    _hash_dict - dictionary between hash of the index and the index in the array

    Methods
    -------
    get_hash - fetch by hash of the tuple
    get_tuple - fetch by tuple (left, hmer_number, hmer_letter, right)
    get_index - directly fetch by index in the array
    """

    def __init__(self, error_model_file: str, n_bins: int = 0):
        """Summary

        Parameters
        ----------
        error_model_file : str
            Description
        n_bins : int, optional
            Description
        """
        if n_bins > 0:
            error_models = [pd.read_hdf(error_model_file, key=f"bin_{b}") for b in range(n_bins)]
            hashed_idcs = [[hash(x) for x in em.index] for em in error_models]
            self.hashed_dict = [
                dict(zip(hashed_idx, range(len(hashed_idx)), strict=False)) for hashed_idx in hashed_idcs
            ]
            del hashed_idcs
            self.error_model = [
                np.array(error_model[[x for x in error_model.columns if isinstance(x, str) and x.startswith("P")]])
                for error_model in error_models
            ]
            for i, _ in enumerate(self.error_model):
                self.error_model[i] = np.concatenate((self.error_model[i], np.zeros((1, self.error_model[i].shape[1]))))
            del error_models
        else:
            error_model = pd.read_hdf(error_model_file, key="error_model_hashed")
            hashed_idx = [hash(x) for x in error_model.index]
            self.hashed_dict = dict(zip(hashed_idx, range(len(hashed_idx)), strict=False))
            del hashed_idx
            self.error_model = np.array(error_model[["P(-1)", "P(0)", "P(+1)"]])
            self.error_model = np.concatenate((self.error_model, np.zeros((1, self.error_model.shape[1]))))
            del error_model

    def hash2idx(self, hash_list: list, bins: np.ndarray | None = None) -> list:
        """Summary

        Parameters
        ----------
        hash_list : list
            Description
        bins : Optional[np.ndarray], optional
            Description

        Returns
        -------
        list
            Description
        """
        if bins is None:
            return [self.hashed_dict.get(x, self.error_model.shape[0] - 1) for x in hash_list]
        return [
            self.hashed_dict[bins[i]].get(hash_list[i], self.error_model[bins[i]].shape[0] - 1)
            for i in range(len(bins))
        ]

    def get_hash(self, tuple_hash: int, hist_bin: int | None = None) -> np.array:
        """Summary

        Parameters
        ----------
        tuple_hash : int
            Hash
        hist_bin : int, optional
            bin

        Returns
        -------
        np.array
            hash
        """
        if hist_bin is None:
            hashed_idx = self.hashed_dict.get(tuple_hash, self.error_model.shape[0] - 1)
            return self.error_model[hashed_idx, :]
        hashed_idx = self.hashed_dict[hist_bin].get(tuple_hash, self.error_model[hist_bin].shape[0] - 1)
        return self.error_model[hist_bin][hashed_idx, :]

    def get_tuple(self, tup: tuple, hist_bin: int | None = None) -> np.array:
        """Summary

        Parameters
        ----------
        tup : tuple
            Description
        hist_bin : int, optional
            Description

        Returns
        -------
        np.array
            Description
        """
        if hist_bin is None:
            hashed_idx = self.hashed_dict.get(hash(tup), self.error_model.shape[0] - 1)
            return self.error_model[hashed_idx, :]
        hashed_idx = self.hashed_dict[hist_bin].get(hash(tup), self.error_model[hist_bin].shape[0] - 1)
        return self.error_model[hist_bin][hashed_idx, :]

    def get_index(self, index_list: np.ndarray, bins: np.ndarray | None = None) -> np.array:
        """Indexing into error model

        Parameters
        ----------
        index_list : np.ndarray
            Description
        bins : Optional[np.ndarray], optional
            Description

        Returns
        -------
        np.array
            Description
        """
        if bins is None:
            return self.error_model[index_list, :]

        output = np.zeros((len(index_list), self.error_model[0].shape[1]))
        bins_set: Counter = Counter(bins)

        for bin_single in bins_set:
            put = bins == bin_single
            output[put, :] = self.error_model[bin_single][index_list[put], :]
        return output



from __future__ import annotations

import argparse
import sys

import numpy as np
from ugbio_core.sorter_utils import read_effective_coverage_from_sorter_json


def parse_args(argv: list[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="sorter_stats_to_mean_coverage", description=run.__doc__)
    parser.add_argument(
        "-i",
        "--sorter-stats-json",
        type=str,
        required=True,
        help="input sorter stats JSON file",
    )
    parser.add_argument(
        "-o",
        "--output-file",
        type=str,
        required=True,
        help="Path of output file",
    )
    return parser.parse_args(argv[1:])


def sorter_stats_to_mean_coverage(sorter_stats_json: str, output_file: str):
    """
    Read mean coverage metric from sorter JSON file and write the up-rounded result to a file.

    Parameters
    ----------
    sorter_stats_json : str
        Path to Sorter statistics JSON file.
    output_file : str
        Path to output file.

    """
    mean_cvg = np.ceil(read_effective_coverage_from_sorter_json(sorter_stats_json)[0])
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(f"{mean_cvg:.0f} ")


def run(argv):
    """Read effective coverage metrics from sorter JSON file and write the up-rounded result to a file."""
    args = parse_args(argv)

    sorter_stats_to_mean_coverage(
        sorter_stats_json=args.sorter_stats_json,
        output_file=args.output_file,
    )


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()



#!/env/python
import argparse
import os
import sys

import pandas as pd
from ugbio_core import picard_metrics_utils


def run(argv: list[str]):
    ap = argparse.ArgumentParser(
        prog="collect_existing_picard_metrics.py",
        description="Collect pipeline metrics in h5 file",
    )
    ap.add_argument(
        "--metric_files",
        nargs="+",
        help="comma seperated list of picard metric files",
        required=False,
    )
    ap.add_argument("--coverage_h5", help="Coverage h5 File", required=False, type=str)
    ap.add_argument("--short_report_h5", help="Short report h5 file", required=False, type=str)
    ap.add_argument("--extended_report_h5", help="Extended report h5 file", required=False, type=str)
    ap.add_argument("--no_gt_report_h5", help="No ground truth report h5 file", required=False, type=str)
    ap.add_argument("--output_h5", help="Aggregated Metrics h5 file", required=False, type=str)
    ap.add_argument("--contamination_stdout", help="Rate of Contamination", required=False, type=str)

    args = ap.parse_args(argv[1:])
    if args.metric_files is not None:
        for metric_file in args.metric_files:
            if os.path.getsize(metric_file) > 0:
                metric_class, stats, histogram = picard_metrics_utils.parse_cvg_metrics(metric_file)
                metric_class = metric_class[metric_class.find("$") + 1 :]
                stats.to_hdf(args.output_h5, key=metric_class, mode="a")
                if histogram is not None:
                    histogram.to_hdf(args.output_h5, key="histogram_" + metric_class, mode="a")

    if args.coverage_h5 is not None:
        cvg_h5_histogram = pd.read_hdf(args.coverage_h5, key="histogram")
        cvg_df = pd.read_hdf(args.coverage_h5, key="stats")
        cvg_df_unstacked = pd.DataFrame(cvg_df.unstack(level=0)).T  # noqa PD010
        cvg_df_unstacked.to_hdf(args.output_h5, key="stats_coverage", mode="a")
        cvg_h5_histogram.to_hdf(args.output_h5, key="histogram_coverage", mode="a")

    if args.short_report_h5 is not None:
        add_h5_to_hdf(args.short_report_h5, args.output_h5, "short_report_")

    if args.extended_report_h5 is not None:
        add_h5_to_hdf(args.extended_report_h5, args.output_h5, "extended_report_")

    if args.contamination_stdout is not None:
        contamination_df = pd.DataFrame(pd.Series(data=[float(args.contamination_stdout)], index=["contamination"])).T
        contamination_df.to_hdf(args.output_h5, key="contamination", mode="a")

    if args.no_gt_report_h5 is not None:
        add_h5_to_hdf(args.no_gt_report_h5, args.output_h5, "no_gt_report_")


def preprocess_columns(df):
    """Handle multiIndex/ hierarchical .h5 - concatenate the columns for using it as single string in JSON."""
    if hasattr(df, "columns"):
        multi_index_seperator = "___"
        if isinstance(df.columns, pd.core.indexes.multi.MultiIndex):
            df.columns = [
                multi_index_seperator.join(col).replace(multi_index_seperator, "") for col in df.columns.to_numpy()
            ]


def add_h5_to_hdf(input_h5_name, output_h5_name, output_report_key_prefix):
    with pd.HDFStore(input_h5_name, "r") as hdf:
        hdf_keys = hdf.keys()
        for report_key in hdf_keys:
            report_h5_pd = pd.read_hdf(input_h5_name, key=report_key)
            report_h5_pd_df = pd.DataFrame(report_h5_pd)
            report_h5_unstacked = pd.DataFrame(
                report_h5_pd_df.unstack(level=list(range(report_h5_pd_df.index.nlevels)))  # noqa PD010
            ).T
            report_h5_unstacked.to_hdf(output_h5_name, key=output_report_key_prefix + report_key, mode="a")


def main():
    run(sys.argv)


if __name__ == "__main__":
    main()
